import unittest
import os
from dilithium_py.dilithium import Dilithium2, Dilithium3, Dilithium5
from dilithium_py.drbg.aes256_ctr_drbg import AES256_CTR_DRBG


def parse_kat_data(data):
    """
    Helper function to parse data from KAT
    file to bytes in a dictionary
    """
    parsed_data = {}
    count_blocks = data.split("\n\n")
    for block in count_blocks[1:-1]:
        block_data = block.split("\n")
        count, seed, mlen, msg, pk, sk, smlen, sm = [
            line.split(" = ")[-1] for line in block_data
        ]
        parsed_data[int(count)] = {
            "seed": bytes.fromhex(seed),
            "msg": bytes.fromhex(msg),
            "mlen": int(mlen),
            "pk": bytes.fromhex(pk),
            "sk": bytes.fromhex(sk),
            "sm": bytes.fromhex(sm),
        }
    return parsed_data


class TestDilithium(unittest.TestCase):
    """
    Test Dilithium for internal
    consistency by generating signatures
    and verifying them!
    """

    def generic_test_dilithium(self, Dilithium):
        msg = b"Signed by dilithium" + os.urandom(16)

        # Perform signature process
        pk, sk = Dilithium.keygen()
        sig = Dilithium.sign(sk, msg)
        check_verify = Dilithium.verify(pk, msg, sig)

        # Generate some fail cases
        pk_bad, _ = Dilithium.keygen()
        check_wrong_pk = Dilithium.verify(pk_bad, msg, sig)
        check_wrong_msg = Dilithium.verify(pk, b"", sig)

        # Check that signature works
        self.assertTrue(check_verify)
        # Check changing the key breaks verify
        self.assertFalse(check_wrong_pk)
        # Check changing the message breaks verify
        self.assertFalse(check_wrong_msg)

    def test_dilithium2(self):
        for _ in range(5):
            self.generic_test_dilithium(Dilithium2)

    def test_dilithium3(self):
        for _ in range(5):
            self.generic_test_dilithium(Dilithium3)

    def test_dilithium5(self):
        for _ in range(5):
            self.generic_test_dilithium(Dilithium5)


class TestDilithiumDRBG(unittest.TestCase):
    """
    Ensure that deterministic DRBG is deterministic!

    Uses AES256 CTR DRBG for randomness.
    Note: requires pycryptodome for AES impl.
    """

    def generic_test_dilithium(self, Dilithium):
        """
        First we generate five pk,sk pairs
        from the same seed and make sure
        they're all the same
        """
        seed = os.urandom(48)
        pk_output = []
        for _ in range(5):
            Dilithium.set_drbg_seed(seed)
            pk, sk = Dilithium.keygen()
            pk_output.append(pk + sk)
        self.assertEqual(len(pk_output), 5)
        self.assertEqual(len(set(pk_output)), 1)

        """
        Now given a fixed keypair make sure
        that all the signatures are the same
        and that they all verify correctly!
        """
        sig_output = []
        seed = os.urandom(48)
        msg = b"Signed by Dilithium" + os.urandom(32)
        pk, sk = Dilithium.keygen()
        for _ in range(5):
            Dilithium.set_drbg_seed(seed)
            sig = Dilithium.sign(sk, msg)
            verify = Dilithium.verify(pk, msg, sig)
            # Check signature worked
            self.assertTrue(verify)
            sig_output.append(sig)

        # Make sure all five signatures are the same
        self.assertEqual(len(sig_output), 5)
        self.assertEqual(len(set(sig_output)), 1)

    def test_dilithium2(self):
        for _ in range(5):
            self.generic_test_dilithium(Dilithium2)

    def test_dilithium3(self):
        for _ in range(5):
            self.generic_test_dilithium(Dilithium3)

    def test_dilithium5(self):
        for _ in range(5):
            self.generic_test_dilithium(Dilithium5)


class TestKnownTestValuesDilithium(unittest.TestCase):
    def generic_test_dilithium(self, Dilithium, file_name):
        entropy_input = bytes([i for i in range(48)])
        drbg = AES256_CTR_DRBG(entropy_input)

        with open(f"assets/{file_name}") as f:
            # extract data from KAT
            kat_data = f.read()
            parsed_data = parse_kat_data(kat_data)

        for count in range(100):
            data = parsed_data[count]

            seed = drbg.random_bytes(48)
            self.assertEqual(data["seed"], seed)

            msg_len = data["mlen"]
            msg = drbg.random_bytes(msg_len)
            self.assertEqual(data["msg"], msg)

            Dilithium.set_drbg_seed(seed)
            pk, sk = Dilithium.keygen()

            # Check that the keygen matches
            self.assertEqual(data["pk"], pk)
            self.assertEqual(data["sk"], sk)

            # Check that the signature matches
            sm_KAT = data["sm"]
            sig_KAT = sm_KAT[:-msg_len]

            # sm_KAT has message as the last mlen bytes
            self.assertEqual(msg, sm_KAT[-msg_len:])

            # Ensure that a generated signature matches
            # the one extracted from the KAT
            sig = Dilithium.sign(sk, msg)
            self.assertEqual(sig, sig_KAT)

            # Finally, make sure that the signature is
            # valid for the message
            verify_KAT = Dilithium.verify(pk, msg, sig)
            self.assertTrue(verify_KAT)

    def test_dilithium2(self):
        self.generic_test_dilithium(Dilithium2, "PQCsignKAT_Dilithium2.rsp")

    def test_dilithium3(self):
        self.generic_test_dilithium(Dilithium3, "PQCsignKAT_Dilithium3.rsp")

    def test_dilithium5(self):
        self.generic_test_dilithium(Dilithium5, "PQCsignKAT_Dilithium5.rsp")


if __name__ == "__main__":
    unittest.main()



import os
from ..modules.modules import Matrix, Module, Vector

try:
    from xoflib import shake256
except ImportError:
    from ..shake.shake_wrapper import shake256


class ML_DSA:
    def __init__(self, parameter_set: dict):
        self.d = parameter_set["d"]
        self.k = parameter_set["k"]
        self.l = parameter_set["l"]
        self.eta = parameter_set["eta"]
        self.tau = parameter_set["tau"]
        self.omega = parameter_set["omega"]
        self.gamma_1 = parameter_set["gamma_1"]
        self.gamma_2 = parameter_set["gamma_2"]
        self.beta = self.tau * self.eta
        self.c_tilde_bytes = parameter_set["c_tilde_bytes"]

        self.M = Module()
        self.R = self.M.ring
        self.oid = parameter_set["oid"] if "oid" in parameter_set else None

        # Use system randomness by default, for deterministic randomness
        # use the method `set_drbg_seed()`
        self.random_bytes = os.urandom

    def set_drbg_seed(self, seed: bytes):
        """
        Change entropy source to a DRBG and seed it with provided value.

        Setting the seed switches the entropy source from :func:`os.urandom()`
        to an AES256 CTR DRBG.

        Used for both deterministic versions of Kyber as well as testing
        alignment with the KAT vectors

        Note:
          currently requires pycryptodome for AES impl.
        """
        try:
            from ..drbg.aes256_ctr_drbg import AES256_CTR_DRBG

            self._drbg = AES256_CTR_DRBG(seed)
            self.random_bytes = self._drbg.random_bytes
        except ImportError as e:  # pragma: no cover
            print(f"Error importing AES from pycryptodome: {e = }")
            raise Warning(
                "Cannot set DRBG seed due to missing dependencies, try installing requirements: pip -r install requirements"
            )

    """
    H() uses Shake256 to hash data to 32 and 64 bytes in a
    few places in the code
    """

    @staticmethod
    def _h(in_bytes: bytes, length: int) -> bytes:
        """
        H: B^*  -> B^*
        """
        return shake256(in_bytes).read(length)

    def _expand_matrix_from_seed(self, rho: bytes) -> Matrix:
        """
        Helper function which generates a element of size
        k x l from a seed `rho`.
        """
        A_data = [[self.R.zero() for _ in range(self.l)] for _ in range(self.k)]
        for i in range(self.k):
            for j in range(self.l):
                A_data[i][j] = self.R.rejection_sample_ntt_poly(rho, i, j)
        return self.M(A_data)

    def _expand_vector_from_seed(self, rho_prime: bytes) -> tuple[Vector, Vector]:
        s1_elements = [
            self.R.rejection_bounded_poly(rho_prime, i, self.eta) for i in range(self.l)
        ]
        s2_elements = [
            self.R.rejection_bounded_poly(rho_prime, i, self.eta)
            for i in range(self.l, self.l + self.k)
        ]

        s1 = self.M.vector(s1_elements)
        s2 = self.M.vector(s2_elements)
        return s1, s2

    def _expand_mask_vector(self, rho: bytes, mu: int) -> Vector:
        elements = [
            self.R.sample_mask_polynomial(rho, i, mu, self.gamma_1)
            for i in range(self.l)
        ]
        return self.M.vector(elements)

    @staticmethod
    def _pack_pk(rho: bytes, t1: Vector) -> bytes:
        return rho + t1.bit_pack_t1()

    def _pack_sk(
        self,
        rho: bytes,
        k: bytes,
        tr: bytes,
        s1: Vector,
        s2: Vector,
        t0: Vector,
    ) -> bytes:
        s1_bytes = s1.bit_pack_s(self.eta)
        s2_bytes = s2.bit_pack_s(self.eta)
        t0_bytes = t0.bit_pack_t0()
        return rho + k + tr + s1_bytes + s2_bytes + t0_bytes

    def _pack_h(self, h: Vector) -> bytes:
        non_zero_positions = [
            [i for i, c in enumerate(poly.coeffs) if c == 1]
            for row in h._data
            for poly in row
        ]
        packed = []
        offsets = []
        for positions in non_zero_positions:
            packed.extend(positions)
            offsets.append(len(packed))

        padding_len = self.omega - offsets[-1]
        packed.extend([0 for _ in range(padding_len)])
        return bytes(packed + offsets)

    def _pack_sig(self, c_tilde: bytes, z: Vector, h: Vector) -> bytes:
        return c_tilde + z.bit_pack_z(self.gamma_1) + self._pack_h(h)

    def _pk_size(self) -> int:
        return 32 + 32 * self.k * 10

    def _unpack_pk(self, pk: bytes) -> tuple[bytes, Vector]:
        if len(pk) != self._pk_size():
            raise ValueError("PK packed bytes is of the wrong length")
        rho, t1_bytes = pk[:32], pk[32:]
        t1 = self.M.bit_unpack_t1(t1_bytes, self.k)
        return rho, t1

    def _sk_size(self) -> int:
        if self.eta == 2:
            s_bytes = 96
        else:
            s_bytes = 128
        s1_len = s_bytes * self.l
        s2_len = s_bytes * self.k
        t0_len = 416 * self.k
        return 2 * 32 + 64 + s1_len + s2_len + t0_len

    def _sig_size(self) -> int:
        if self.gamma_1 == 131072:
            s_bytes = 18
        else:
            assert self.gamma_1 == 524288
            s_bytes = 20
        return self.c_tilde_bytes + self.l * 32 * s_bytes + self.omega + self.k

    def _unpack_sk(
        self, sk: bytes
    ) -> tuple[bytes, bytes, bytes, Vector, Vector, Vector]:
        if self.eta == 2:
            s_bytes = 96
        else:
            s_bytes = 128
        s1_len = s_bytes * self.l
        s2_len = s_bytes * self.k
        t0_len = 416 * self.k
        if len(sk) != self._sk_size():
            raise ValueError("sk packed bytes is of the wrong length")

        # Split bytes between seeds and vectors
        sk_seed_bytes, sk_vec_bytes = sk[:128], sk[128:]

        # Unpack seed bytes
        rho, k, tr = (
            sk_seed_bytes[:32],
            sk_seed_bytes[32:64],
            sk_seed_bytes[64:128],
        )

        # Unpack vector bytes
        s1_bytes = sk_vec_bytes[:s1_len]
        s2_bytes = sk_vec_bytes[s1_len : s1_len + s2_len]
        t0_bytes = sk_vec_bytes[-t0_len:]

        # Unpack bytes to vectors
        s1 = self.M.bit_unpack_s(s1_bytes, self.l, self.eta)
        s2 = self.M.bit_unpack_s(s2_bytes, self.k, self.eta)
        t0 = self.M.bit_unpack_t0(t0_bytes, self.k)

        return rho, k, tr, s1, s2, t0

    def _unpack_h(self, h_bytes: bytes) -> Vector:
        offsets = [0] + list(h_bytes[-self.k :])

        # ensure offsets are monotonic increasing
        if any(offsets[i] > offsets[i + 1] for i in range(len(offsets) - 1)):
            raise ValueError("offsets in h_bytes are not monotonically increasing")

        # ensure offset[-1] is smaller than the length of h_bytes
        if offsets[-1] > self.omega:
            raise ValueError("accumulate offset of hints exceeds omega")

        # ensure zero fields are all zeros
        if any(b != 0 for b in h_bytes[offsets[-1] : self.omega]):
            raise ValueError("non-zero fields in h_bytes are not all zeros")

        non_zero_positions = [
            list(h_bytes[offsets[i] : offsets[i + 1]]) for i in range(self.k)
        ]

        vector_coeffs = []
        for poly_non_zero in non_zero_positions:
            coeffs = [0 for _ in range(256)]
            for i, non_zero in enumerate(poly_non_zero):
                if i > 0 and non_zero < poly_non_zero[i - 1]:
                    raise ValueError(
                        "non-zero positions in h_bytes are not monotonically increasing"
                    )
                coeffs[non_zero] = 1
            vector_coeffs.append(self.R(coeffs))

        return self.M.vector(vector_coeffs)

    def _unpack_sig(self, sig: bytes) -> tuple[bytes, Vector, Vector]:
        if len(sig) != self._sig_size():
            raise ValueError("Incorrect signature size")
        c_tilde = sig[: self.c_tilde_bytes]
        z_bytes = sig[self.c_tilde_bytes : -(self.k + self.omega)]
        h_bytes = sig[-(self.k + self.omega) :]

        z = self.M.bit_unpack_z(z_bytes, self.l, self.gamma_1)
        h = self._unpack_h(h_bytes)

        return c_tilde, z, h

    def _keygen_internal(self, zeta: bytes) -> tuple[bytes, bytes]:
        """
        Generates a public-private key pair from a seed following
        Algorithm 6 (FIPS 204)
        """
        # Expand with an XOF (SHAKE256)
        seed_domain_sep = zeta + bytes([self.k]) + bytes([self.l])
        seed_bytes = self._h(seed_domain_sep, 128)

        # Split bytes into suitable chunks
        rho, rho_prime, K = seed_bytes[:32], seed_bytes[32:96], seed_bytes[96:]

        # Generate matrix A ∈ R^(kxl) in the NTT domain
        A_hat = self._expand_matrix_from_seed(rho)

        # Generate the error vectors s1 ∈ R^l, s2 ∈ R^k
        s1, s2 = self._expand_vector_from_seed(rho_prime)

        # Matrix multiplication
        s1_hat = s1.to_ntt()
        t = (A_hat @ s1_hat).from_ntt() + s2

        t1, t0 = t.power_2_round(self.d)

        # Pack up the bytes
        pk = self._pack_pk(rho, t1)
        tr = self._h(pk, 64)
        sk = self._pack_sk(rho, K, tr, s1, s2, t0)

        return pk, sk

    def _sign_internal(
        self, sk: bytes, m: bytes, rnd: bytes, external_mu: bool = False
    ) -> bytes:
        """
        Deterministic algorithm to generate a signature for a formatted message
        M' following Algorithm 7 (FIPS 204)

        When `external_mu` is `True`, the message `m` is interpreted instead as
        the pre-hashed message `mu = prehash_external_mu()`
        """
        # unpack the secret key
        rho, k, tr, s1, s2, t0 = self._unpack_sk(sk)

        # Precompute NTT representation
        s1_hat = s1.to_ntt()
        s2_hat = s2.to_ntt()
        t0_hat = t0.to_ntt()

        # Generate matrix A ∈ R^(kxl) in the NTT domain
        A_hat = self._expand_matrix_from_seed(rho)

        # Set seeds and nonce (kappa)
        if external_mu:
            mu = m
        else:
            mu = self._h(tr + m, 64)
        rho_prime = self._h(k + rnd + mu, 64)

        kappa = 0
        alpha = self.gamma_2 << 1
        while True:
            y = self._expand_mask_vector(rho_prime, kappa)
            y_hat = y.to_ntt()
            w = (A_hat @ y_hat).from_ntt()

            # increment the nonce
            kappa += self.l

            # NOTE: there is an optimisation possible where both the high and
            # low bits of w are extracted here, which speeds up some checks
            # below and requires the use of make_hint_optimised() -- to see the
            # implementation of this, look at the signing algorithm for
            # dilithium. We include this slower version to mirror the FIPS 204
            # document precisely.
            # Extract out only the high bits
            w1 = w.high_bits(alpha)

            # Create challenge polynomial
            w1_bytes = w1.bit_pack_w(self.gamma_2)
            c_tilde = self._h(mu + w1_bytes, self.c_tilde_bytes)
            c = self.R.sample_in_ball(c_tilde, self.tau)
            c_hat = c.to_ntt()

            # NOTE: unlike FIPS 204 we start again as soon as a vector
            # fails the norm bound to reduce any unneeded computations.
            c_s1 = s1_hat.scale(c_hat).from_ntt()
            z = y + c_s1
            if z.check_norm_bound(self.gamma_1 - self.beta):
                continue

            c_s2 = s2_hat.scale(c_hat).from_ntt()
            r0 = (w - c_s2).low_bits(alpha)
            if r0.check_norm_bound(self.gamma_2 - self.beta):
                continue

            c_t0 = t0_hat.scale(c_hat).from_ntt()
            if c_t0.check_norm_bound(self.gamma_2):
                continue

            h = (-c_t0).make_hint(w - c_s2 + c_t0, alpha)
            if h.sum_hint() > self.omega:
                continue

            return self._pack_sig(c_tilde, z, h)

    def _verify_internal(self, pk: bytes, m: bytes, sig: bytes) -> bool:
        """
        Internal function to verify a signature sigma for a formatted message M'
        following Algorithm 8 (FIPS 204)
        """
        rho, t1 = self._unpack_pk(pk)
        try:
            c_tilde, z, h = self._unpack_sig(sig)
        except ValueError:
            return False

        if h.sum_hint() > self.omega:
            return False

        if z.check_norm_bound(self.gamma_1 - self.beta):
            return False

        A_hat = self._expand_matrix_from_seed(rho)

        tr = self._h(pk, 64)
        mu = self._h(tr + m, 64)
        c = self.R.sample_in_ball(c_tilde, self.tau)

        # Convert to NTT for computation
        c = c.to_ntt()
        z = z.to_ntt()

        t1 = t1.scale(1 << self.d)
        t1 = t1.to_ntt()

        Az_minus_ct1 = (A_hat @ z) - t1.scale(c)
        Az_minus_ct1 = Az_minus_ct1.from_ntt()

        w_prime = h.use_hint(Az_minus_ct1, 2 * self.gamma_2)
        w_prime_bytes = w_prime.bit_pack_w(self.gamma_2)

        return c_tilde == self._h(mu + w_prime_bytes, self.c_tilde_bytes)

    def keygen(self) -> tuple[bytes, bytes]:
        """
        Generates a public-private key pair following
        Algorithm 1 (FIPS 204)
        """
        zeta = self.random_bytes(32)
        pk, sk = self._keygen_internal(zeta)
        return (pk, sk)

    def key_derive(self, seed: bytes) -> tuple[bytes, bytes]:
        """
        Derive a verification key and corresponding signing key
        following the approach from Section 6.1 (FIPS 204)
        with storage of the ``seed`` value for later expansion.

        ``seed`` is a byte-encoded concatenation of the ``xi`` value.

        :return: Tuple with verification key and signing key.
        :rtype: tuple(bytes, bytes)
        """
        if len(seed) != 32:
            raise ValueError("The seed must be 32 bytes long")

        pk, sk = self._keygen_internal(seed)
        return (pk, sk)

    def sign(
        self, sk: bytes, m: bytes, ctx: bytes = b"", deterministic: bool = False
    ) -> bytes:
        """
        Generates an ML-DSA signature following
        Algorithm 2 (FIPS 204)
        """
        if len(ctx) > 255:
            raise ValueError(
                f"ctx bytes must have length at most 255, ctx has length {len(ctx) = }"
            )

        if deterministic:
            rnd = bytes([0] * 32)
        else:
            rnd = self.random_bytes(32)

        # Format the message using the context
        m_prime = bytes([0]) + bytes([len(ctx)]) + ctx + m

        # Compute the signature of m_prime
        sig_bytes = self._sign_internal(sk, m_prime, rnd)
        return sig_bytes

    def verify(self, pk: bytes, m: bytes, sig: bytes, ctx: bytes = b"") -> bool:
        """
        Verifies a signature sigma for a message M following
        Algorithm 3 (FIPS 204)
        """
        if len(ctx) > 255:
            raise ValueError(
                f"ctx bytes must have length at most 255, ctx has length {len(ctx) = }"
            )

        # Format the message using the context
        m_prime = bytes([0]) + bytes([len(ctx)]) + ctx + m

        return self._verify_internal(pk, m_prime, sig)

    """
    The following additional function follows an outline from:
        https://github.com/aws/aws-lc/pull/2142
    which computes pk_bytes when only the sk_bytes are known.
    """

    def pk_from_sk(self, sk: bytes) -> bytes:
        """
        Given the packed representation of a ML-DSA secret key,
        compute the corresponding packed public key bytes.
        """
        # First unpack the secret key
        rho, _, tr, s1, s2, _ = self._unpack_sk(sk)

        # Compute the matrix A from rho in NTT form
        A_hat = self._expand_matrix_from_seed(rho)

        # Convert s1 to NTT form
        s1_hat = s1.to_ntt()

        # Compute the polynomial t, we have the lower bits t0,
        # but we need the higher bits t1 for the public key
        t = (A_hat @ s1_hat).from_ntt() + s2
        t1, _ = t.power_2_round(self.d)

        # The packed public key is made from rho || t1
        pk = self._pack_pk(rho, t1)

        # Ensure the public key matches the hash within the secret key
        if tr != self._h(pk, 64):
            raise ValueError("malformed secret key")

        return pk

    """
    The following external mu functions are not in FIPS 204, but are in
    Appendix D of the following IETF draft and are included for experimentation
    for researchers and engineers

    https://datatracker.ietf.org/doc/html/draft-ietf-lamps-dilithium-certificates-07
    """

    def prehash_external_mu(self, pk: bytes, m: bytes, ctx: bytes = b"") -> bytes:
        """
        Prehash the message `m` with context `ctx` together with
        the public key. For use with `sign_external_mu()`
        """
        # Ensure the length of the context is as expected
        if len(ctx) > 255:
            raise ValueError(
                f"ctx bytes must have length at most 255, ctx has length {len(ctx) = }"
            )
        if len(pk) != self._pk_size():
            raise ValueError(
                f"Public key size doesn't match this ML-DSA object parameters,"
                f"received {len(pk) = }, expected: {self._pk_size()}"
            )

        # Format the message using the context
        m_prime = bytes([0]) + bytes([len(ctx)]) + ctx + m

        # Compute mu by hashing the public key into the message
        tr = self._h(pk, 64)
        mu = self._h(tr + m_prime, 64)

        return mu

    def sign_external_mu(
        self, sk: bytes, mu: bytes, deterministic: bool = False
    ) -> bytes:
        """
        Generates an ML-DSA signature of a message given the prehash
        mu = H(H(pk), M')
        """
        # Ensure the length of the context is as expected
        if len(mu) != 64:
            raise ValueError(
                f"mu bytes must have length 64, mu has length {len(mu) = }"
            )

        if deterministic:
            rnd = bytes([0] * 32)
        else:
            rnd = self.random_bytes(32)

        # Compute the signature given external mu, we set the external_mu
        # to True
        sig = self._sign_internal(sk, mu, rnd, external_mu=True)
        return sig



import unittest
import os
import json
from dilithium_py.ml_dsa import ML_DSA_44, ML_DSA_65, ML_DSA_87


class TestMLDSA(unittest.TestCase):
    """
    Test ML DSA for internal consistency by generating signatures
    and verifying them!
    """

    def generic_test_ml_dsa(self, ML_DSA, count=5):
        for _ in range(count):
            msg = b"Signed by ML_DSA" + os.urandom(16)
            ctx = os.urandom(128)

            # Perform signature process
            pk, sk = ML_DSA.keygen()
            sig = ML_DSA.sign(sk, msg, ctx=ctx)
            check_verify = ML_DSA.verify(pk, msg, sig, ctx=ctx)

            check_short_verify = ML_DSA.verify(pk, msg, sig[:-1], ctx=ctx)
            check_long_verify = ML_DSA.verify(pk, msg, sig + b"\x00", ctx=ctx)
            check_empty_verify = ML_DSA.verify(pk, msg, b"", ctx=ctx)

            # Sign with external_mu instead
            external_mu = ML_DSA.prehash_external_mu(pk, msg, ctx=ctx)
            sig_external_mu = ML_DSA.sign_external_mu(sk, external_mu)
            check_external_mu = ML_DSA.verify(pk, msg, sig_external_mu, ctx=ctx)

            # Generate some fail cases
            pk_bad, _ = ML_DSA.keygen()
            check_wrong_pk = ML_DSA.verify(pk_bad, msg, sig, ctx=ctx)
            check_wrong_msg = ML_DSA.verify(pk, b"", sig, ctx=ctx)
            check_no_ctx = ML_DSA.verify(pk, msg, sig)

            # Generate the public key directly from the secret key
            recovered_pk = ML_DSA.pk_from_sk(sk)

            # Check that recovering the pk works
            self.assertEqual(pk, recovered_pk)

            # Check that signature works
            self.assertTrue(check_verify)

            # Check that too short signature is rejected
            self.assertFalse(check_short_verify)

            # Check that too short signature is rejected
            self.assertFalse(check_long_verify)

            # Check that empty signature is rejected
            self.assertFalse(check_empty_verify)

            # Check that external_mu also works
            self.assertTrue(check_external_mu)

            # Check changing the key breaks verify
            self.assertFalse(check_wrong_pk)

            # Check changing the message breaks verify
            self.assertFalse(check_wrong_msg)

            # Check removing the context breaks verify
            self.assertFalse(check_no_ctx)

    def test_ml_dsa_44(self):
        self.generic_test_ml_dsa(ML_DSA_44)

    def test_ml_dsa_65(self):
        self.generic_test_ml_dsa(ML_DSA_65)

    def test_ml_dsa_87(self):
        self.generic_test_ml_dsa(ML_DSA_87)


class TestMLDSADeterministic(unittest.TestCase):
    """
    Test ML DSA for internal consistency by generating signatures
    and verifying them!
    """

    def generic_test_ml_dsa(self, ML_DSA, count=5):
        for _ in range(count):
            msg = b"Signed by ML_DSA" + os.urandom(16)
            ctx = os.urandom(128)

            # Perform signature process
            pk, sk = ML_DSA.keygen()
            sig = ML_DSA.sign(sk, msg, ctx=ctx, deterministic=True)
            check_verify = ML_DSA.verify(pk, msg, sig, ctx=ctx)

            # Sign with external_mu instead
            external_mu = ML_DSA.prehash_external_mu(pk, msg, ctx=ctx)
            sig_external_mu = ML_DSA.sign_external_mu(
                sk, external_mu, deterministic=True
            )
            check_external_mu = ML_DSA.verify(pk, msg, sig_external_mu, ctx=ctx)

            # Generate some fail cases
            pk_bad, _ = ML_DSA.keygen()
            check_wrong_pk = ML_DSA.verify(pk_bad, msg, sig, ctx=ctx)
            check_wrong_msg = ML_DSA.verify(pk, b"", sig, ctx=ctx)
            check_no_ctx = ML_DSA.verify(pk, msg, sig)

            # Check that signature works
            self.assertTrue(check_verify)

            # Check that external_mu also works
            self.assertTrue(check_external_mu)

            # Check changing the key breaks verify
            self.assertFalse(check_wrong_pk)

            # Check changing the message breaks verify
            self.assertFalse(check_wrong_msg)

            # Check removing the context breaks verify
            self.assertFalse(check_no_ctx)

    def test_ml_dsa_44(self):
        self.generic_test_ml_dsa(ML_DSA_44)

    def test_ml_dsa_65(self):
        self.generic_test_ml_dsa(ML_DSA_65)

    def test_ml_dsa_87(self):
        self.generic_test_ml_dsa(ML_DSA_87)

    def test_derive_with_wrong_seed_length(self):
        with self.assertRaises(ValueError) as e:
            ML_DSA_44.key_derive(bytes(range(31)))

        self.assertIn("seed must be 32 bytes long", str(e.exception))

    # test vectors copied from
    # https://datatracker.ietf.org/doc/html/draft-ietf-lamps-dilithium-certificates-07
    def test_derive_from_seed_ML_DSA_44(self):
        pk, _ = ML_DSA_44.key_derive(bytes(range(32)))

        exp_pk = bytes.fromhex(
            """
                d7b2b47254aae0db45e7930d4a98d2c97d8f1397d1789dafa17024b316e9bec9
                4fc9946d42f19b79a7413bbaa33e7149cb42ed5115693ac041facb988adeb5fe
                0e1d8631184995b592c397d2294e2e14f90aa414ba3826899ac43f4cccacbc26
                e9a832b95118d5cb433cbef9660b00138e0817f61e762ca274c36ad554eb22aa
                c1162e4ab01acba1e38c4efd8f80b65b333d0f72e55dfe71ce9c1ebb9889e7c5
                6106c0fd73803a2aecfeafded7aa3cb2ceda54d12bd8cd36a78cf975943b47ab
                d25e880ac452e5742ed1e8d1a82afa86e590c758c15ae4d2840d92bca1a5090f
                40496597fca7d8b9513f1a1bda6e950aaa98de467507d4a4f5a4f0599216582c
                3572f62eda8905ab3581670c4a02777a33e0ca7295fd8f4ff6d1a0a3a7683d65
                f5f5f7fc60da023e826c5f92144c02f7d1ba1075987553ea9367fcd76d990b7f
                a99cd45afdb8836d43e459f5187df058479709a01ea6835935fa70460990cd3d
                c1ba401ba94bab1dde41ac67ab3319dcaca06048d4c4eef27ee13a9c17d0538f
                430f2d642dc2415660de78877d8d8abc72523978c042e4285f4319846c441262
                42976844c10e556ba215b5a719e59d0c6b2a96d39859071fdcc2cde7524a7bed
                ae54e85b318e854e8fe2b2f3edfac9719128270aafd1e5044c3a4fdafd9ff31f
                90784b8e8e4596144a0daf586511d3d9962b9ea95af197b4e5fc60f2b1ed15de
                3a5bef5f89bdc79d91051d9b2816e74fa54531efdc1cbe74d448857f476bcd58
                f21c0b653b3b76a4e076a6559a302718555cc63f74859aabab925f023861ca8c
                d0f7badb2871f67d55326d7451135ad45f4a1ba69118fbb2c8a30eec9392ef3f
                977066c9add5c710cc647b1514d217d958c7017c3e90fd20c04e674b90486e93
                70a31a001d32f473979e4906749e7e477fa0b74508f8a5f2378312b83c25bd38
                8ca0b0fff7478baf42b71667edaac97c46b129643e586e5b055a0c211946d4f3
                6e675bed5860fa042a315d9826164d6a9237c35a5fbf495490a5bd4df248b95c
                4aae7784b605673166ac4245b5b4b082a09e9323e62f2078c5b76783446defd7
                36ad3a3702d49b089844900a61833397bc4419b30d7a97a0b387c1911474c4d4
                1b53e32a977acb6f0ea75db65bb39e59e701e76957def6f2d44559c31a77122b
                5204e3b5c219f1688b14ed0bc0b801b3e6e82dcd43e9c0e9f41744cd9815bd1b
                c8820d8bb123f04facd1b1b685dd5a2b1b8dbbf3ed933670f095a180b4f192d0
                8b10b8fabbdfcc2b24518e32eea0a5e0c904ca844780083f3b0cd2d0b8b6af67
                bc355b9494025dc7b0a78fa80e3a2dbfeb51328851d6078198e9493651ae787e
                c0251f922ba30e9f51df62a6d72784cf3dd205393176dfa324a512bd94970a36
                dd34a514a86791f0eb36f0145b09ab64651b4a0313b299611a2a1c4889162759
                8768a3114060ba4443486df51522a1ce88b30985c216f8e6ed178dd567b304a0
                d4cafba882a28342f17a9aa26ae58db630083d2c358fdf566c3f5d62a428567b
                c9ea8ce95caa0f35474b0bfa8f339a250ab4dfcf2083be8eefbc1055e18fe153
                70eecb260566d83ff06b211aaec43ca29b54ccd00f8815a2465ef0b46515cc7e
                41f3124f09efff739309ab58b29a1459a00bce5038e938c9678f72eb0e4ee5fd
                aae66d9f8573fc97fc42b4959f4bf8b61d78433e86b0335d6e9191c4d8bf487b
                3905c108cfd6ac24b0ceb7dcb7cf51f84d0ed687b95eaeb1c533c06f0d97023d
                92a70825837b59ba6cb7d4e56b0a87c203862ae8f315ba5925e8edefa679369a
                2202766151f16a965f9f81ece76cc070b55869e4db9784cf05c830b3242c8312
            """
        )

        self.assertEqual(len(pk), len(exp_pk))
        self.assertEqual(pk, exp_pk)

    # test vectors copied from
    # https://datatracker.ietf.org/doc/html/draft-ietf-lamps-dilithium-certificates-07
    def test_derive_from_seed_ML_DSA_65(self):
        pk, _ = ML_DSA_65.key_derive(bytes(range(32)))

        exp_pk = bytes.fromhex(
            """
            48683d91978e31eb3dddb8b0473482d2b88a5f625949fd8f58a561e696bd4c27
            d05b38dbb2edf01e664efd81be1ea893688ce68aa2d51c5958f8bbc6eb4e89ee
            67d2c0320954d57212cac7229ff1d6eaf03928bd51511f8d88d847736c7de273
            0d5978e5410713160978867711bf5539a0bfc4c350c2be572baf0ee2e2fb16cc
            fea08028d99ac49aebb75937ddce111cdab62fff3cea8ba2233d1e56fbc5c5a1
            e726de63fadd2af016b119177fa3d971a2d9277173fce55b67745af0b7c21d59
            7dbeb93e6a32f341c49a5a8be9e825088d1f2aa45155d6c8ae15367e4eb003b8
            fdf7851071949739f9fff09023eaf45104d2a84a45906eed4671a44dc28d2798
            7bb55df69e9e8561f61a80a72699503865fed9b7ee72a8e17a19c408144f4b29
            afef7031c3a6d8571610b42c9f421245a88f197e16812b031159b65b9687e5b3
            e934c5225ae98a79ba73d2b399d73510effad19e53b8450f0ba8fce1012fd98d
            260a74aaaa13fae249a006b1c34f5ba0b882f26378222fb36f2283c243f0ffeb
            5f1bb414a0a70d55e3d40a56b6cbc88ae1f03b7b2882d98deea28e145c9dedfd
            8eaf1cef2ed94a8b050f8964f46d1ea0d0c2a43e0dda6182adbf4f6ed175b674
            2257859bf22f3a417ecf1f9d89317b5e539d587af16b9e1313e04514ffa64ba8
            b3ff2b8321f8811cb3fb022c8f644e70a4b80a2fbfee604abb7379091ea8e6c5
            c74dfc0283666b40c0793870028204a136bf5da9568eb798d349038bdb0c11e0
            3445e7847cb5069c75cf28ac601c7799d958210ddbcb226e51afef9f1de47b07
            3873d6d3f97456bede085082e74a298b2cd48f4b3093155f366c8fa601c6af85
            8dfa32c08491b2a29887f90335949a5d6edaa679882a3a95d6bf6d970a221f4b
            9d3d8cbf384af81aac95e2b3294e04789ac83727a5dc04559f96af41d8a05351
            6feeeebc52746eb6ab2819e09108710d835f011fa63065872ad334d5cdffb2b2
            310507e92fc993ae317da97f4f309cdaf0f67ed99d90215576083849f953b246
            d7fedb3fdb67679850a5ad404e64147fb7cf4f6aeddd05afb4b834968d1fe880
            14960dce5d942236526e12a478d69e5fbe6970310b308c06845018cfc7b2ab43
            0a13a6b1ac7bb02cccbb3d911ac2f11068613fbe029bfdce02cf5cd38950ed72
            c83944edfbc75615af87f864c051f3c55456c5412863a40c06d1dab562bdff05
            71b8d3c3917bbd300880bba5e998239b95fa91b7d6416d4f398b3adbcd30983e
            d3592b4d9ef7d4236fd00f50d98aa53a235ac4172720f77d96172672980cfe8f
            f7a5a702783edc2ba31b2259015a112fc7f468a9c2f9464039002d30ef678b4c
            b798bc116216bf7a9a7c18ba03b7b58fd07515d3115049d3614be7a07e744300
            750df1d2c58753389059eafc3d785ccdd31c07648bedc03a5c3b8ad46d064d59
            c13d57374729fc4e295362e2a5191204530428bc1522afa28ff5fe1655e304ca
            5bc8c27ad0e0c6a39dd4df28956c14b38cc93682cefe402bbd5e82d29c464e44
            eb5d37b48fc568dfe0cc6e8e16baea05e5135590f19294e73e8367b0216dbb81
            5030b9de55913f08039c42351c59e5515dd5af8e089a15e625e8f6dee639386c
            46497d7a263288774de581a7de9629b41b4424141f978fb8331208efdec3c6e0
            de39bc57063f3dcd6c470373c08891ea29cbc7cc6d6483b8889083ace86aa7b5
            1b1c2cfe6e2ad18d97ce36fbc56ea42fae97e6a7ac114864478c366df1ebb1e7
            b11a9098504fd5975bdf1f49dc70002b63c1739a9d263fbad4073f6a9f6c2b8a
            f4b4c332a103a0cffa5deeb2d062ca3c215fd360026be7c5164f4a4424ef7494
            8804d66f46487732c8202c795478647b4ea71d627c086024cca354a41f0877b3
            8f19b3774ad2095c8da53b069e21c76ae2d2007e16719ed40080d334f7da52e9
            f5a5990439caf083a95b833f02ad10a08c1a6d0f260c007285bd4a2f47703a5a
            ef465287d253b18ac22514316210ff566814b10f87a293d6f199d3c3959990d0
            c1268b4f50d5f9fcefbbf237bd0c28b80182d6659741f14f10bfbb21bba12ab6
            20aa2396f56c0686b4ea9017990224216b2fe8ad76c4a9148eef9a86a3635a6a
            a77bc1dcfb6fba59a77dfda9b7530dc0ca8648c8d973738e01bab8f08b4905e8
            4aa4641bd602410cd97520265f2f231f2b35e15eb2fa04d2bd94d5a77abaf1e0
            e161010a990087f5b46ea988b2bc0512fda0fa923dadd6c45c5301d094836732
            65b5ab2e10f4ba520f6bbad564a5c3d5e27bdb080f7d20e13296a3181954c39c
            649c943ebe17df5c1f7aae0a8fe126c477585a5d4d648a0d008b6af5e8cd31be
            69a9296d4f3fd25ed86f221e4b93f65f5929967533624b9235750c30707550b5
            8536d109a7131c5a5bbe4a5715567c12534aec7660761eebb9fae2891c774589
            b80e566ad557ddef7367196b7227ea9870ef09ddfec79d6b9319a6879b5205d7
            6bf7aba5acf33afb59d17fc54e68383d6be5a08e9b66da53dcde008bb294b858
            2bd132cdcc49959fdbc21e52721880c8ad0352c79f03a43bbd84c4cdfdc6c529
            005e1e7cd9a349a7168a35569ba5dea818968d5a91466bd6e64e20bf62417198
            afc4e81c28dd77ed4028232398b52fbde86bc84f475b9016710ce2aabc11a06b
            4dbac901ec16cf365ca3f2d53813948a693a0f93e79c46ca5d5a6dca3d28ca50
            ad18bd13fca55059dd9b185f79f9c47196a4e81b2104bc460a051e02f2e8444f
        """
        )

        self.assertEqual(len(pk), len(exp_pk))
        self.assertEqual(pk, exp_pk)

    # test vectors copied from
    # https://datatracker.ietf.org/doc/html/draft-ietf-lamps-dilithium-certificates-07
    def test_derive_from_seed_ML_DSA_87(self):
        pk, _ = ML_DSA_87.key_derive(bytes(range(32)))

        exp_pk = bytes.fromhex(
            """
            9792bcec2f2430686a82fccf3c2f5ff665e771d7ab41b90258cfa7e90ec97124
            a73b323b9ba21ab64d767c433f5a521effe18f86e46a188952c4467e048b729e
            7fc4d115e7e48da1896d5fe119b10dcddef62cb307954074b42336e52836de61
            da941f8d37ea68ac8106fabe19070679af6008537120f70793b8ea9cc0e6e7b7
            b4c9a5c7421c60f24451ba1e933db1a2ee16c79559f21b3d1b8305850aa42afb
            b13f1f4d5b9f4835f9d87dfceb162d0ef4a7fdc4cba1743cd1c87bb4967da16c
            c8764b6569df8ee5bdcbffe9a4e05748e6fdf225af9e4eeb7773b62e8f85f9b5
            6b548945551844fbd89806a4ac369bed2d256100f688a6ad5e0a709826dc4449
            e91e23c5506e642361ef5a313712f79bc4b3186861ca85a4bab17e7f943d1b8a
            333aa3ae7ce16b440d6018f9e04daf5725c7f1a93fad1a5a27b67895bd249aa9
            1685de20af32c8b7e268c7f96877d0c85001135a4f0a8f1b8264fa6ebe5a349d
            8aecad1a16299ccf2fd9c7b85bace2ced3aa1276ba61ee78ed7e5ca5b67cdd45
            8a9354030e6abbbabf56a0a2316fec9dba83b51d42fd3167f1e0f90855d5c665
            09b210265dc1e54ec44b43ba7cf9aef118b44d80912ce75166a6651e116cebe4
            9229a7062c09931f71abd2293f76f7efc3215ba97800037e58e470bdbbb43c1b
            0439eaf79c54d93b44aac9efe9fbe151874cfb2a64cbee28cc4c0fe7775e5d87
            0f1c02e5b2e3c5004c995f24c9b779cb753a277d0e71fd425eb6bc2ca56ce129
            db51f70740f31e63976b50c7312e9797d78c5b1ac24a5fa347cc916e0a83f5c3
            b675cd30b81e3fa10b93444e07397571cce98b28da51db9056bc728c5b0b1181
            e2fbd387b4c79ab1a5fefece37167af772ddad14eb4c3982da5a59d0e9eb173e
            c6315091170027a3ab5ef6aa129cb8585727b9358a28501d713a72f3f1db3171
            4286f9b6408013af06045d75592fc0b7dd47c73ed9c75b11e9d7c69f7cadfc32
            80a9062c5273c43be1c34f87448864cea7b5c97d6d32f59bd5f25384653bb5c4
            faa45bea8b89402843e645b6b9269e2bd988ddacb033328ffb060450f7df0800
            53e6969b251e875ecec32cfc592840d69ab69a75e06b379c535d95266b082f4f
            09c93162b33b0d9f7307a4eaaa52104437fed66f8ee3eabbd45d67b25a8133f4
            96468b52baffdbfad93eef1a9818b5e42ec722788a3d8d3529fc777d2ba57080
            1dfae01ec88302837c1fb9e0355727645ee1046c3f915f6ae82dad4fb6b0356a
            46518ffc834155c3b4fe6dafa6cc8a5ccf53c73a0849d8d44f7dcf72754e70e1
            b7dfb447bb4ef49d1a718f6171bbce200950e0ce926106b151a3e871d5ce4973
            1bd6650a9b0ca972da1c5f136d44820ea6383c08f3b384cf2338e789c513f618
            cc5694a6f0cee104511e1ed7c5f23a1ebfd8a0db8424553240156dbf622831b0
            c643d1c551b6f3f7a98d29b85c2de05a65fa615eee16495bd90737672115b53e
            91c5d90028cf3f1a93953a153de53b44084e9ccff6b736693926daefebb2d77a
            a5ad689b92f31686669df16d1715cc58f7a2cfb72dd1a51e92f825993a74022b
            e7e9eb6054654457094d14928f20215e7b222ac56b51adbec8d8bdb6983979a7
            e3a21b44b5d1518ca97d0b5195f51ed6a24350c89747e1edea51b448e3e91470
            54ce927873c90db394d86888e07dff177593d6f79e152302204aeb03be2386af
            3e24078bd028b1689f5e147c9f452c8ceb02ec59cc9db63a03576ceeafe98239
            023897da0236630a53c0de7f435a19869792fab36e7b9e635760f09069e6432e
            700035ac2a02879fff0a1e1bec522047193d94eb5df1efd53eea1144ca789408
            52f5ec9727904b366ede4f5e2d331fad5fc282ea2c47e923142771c3dd75a873
            57487def99e5f18e9d9ed623c175d02888c51f82c07a80d54716b3c3c2bdbe2e
            9f0a9bbaaebeb4d52936876406f5c00e8e4bbd0a5ec05797e6207c5ab6c88f1a
            688421bd05a114f4d7de2ac241fa0e8bedff47f762ddcbeaa91004f8d31e8509
            5c81054994ad3826e344ba96040810fc0b2ad1de48cfade002c62e5a49a0731a
            b38344bc1636df16bf607d56855e56d684003c718e4bad9e5a099979fcddeeb1
            c4a7776cd37a3417cb0e184e29ef9bc0e87475ba663be09e00ab562eb7c0f716
            5f969a9b42414198ccf1bff2a2c8d689a414ece7662927665689e94db961ebae
            c5615cbc1a7895c6851ac961432ff1118d4607d32ef9dc732d51333be4b4d0e3
            0ddea784eca8be47e741be9c19631dc470a52ef4dc13a4f3633fd434d787c170
            977b417df598e1d0dde506bb71d6f0bc17ec70e3b03cdc1965cb36993f633b04
            72e50d0923ac6c66fdf1d3e6459cc121f0f5f94d09e9dbcf5d690e23233838a0
            bacb7c638d1b2650a4308cd171b6855126d1da672a6ed85a8d78c286fb56f4ab
            3d21497528045c63262c8a42af2f9802c53b7bb8be28e78fe0b5ce45fbb7a1af
            1a3b28a8d94b7890e3c882e39bc98e9f0ad76025bf0dd2f00298e7141a226b3d
            7cee414f604d1e0ba54d11d5fe58bccea6ad77ad2e8c1caacf32459014b7b910
            01b1efa8ad172a523fb8e365b577121bf9fd88a2c60c21e821d7b6acb47a5a99
            5e40caced5c223b8fe6de5e18e9d2e5893aefebb7aae7ff1a146260e2f110e93
            9528213a0025a38ec79aabc861b25ebc509a4674c132aaacb7e0146f14efd11c
            fcaf4caa4f775a716ce325e0a435a4d349d720bcf137450afc45046fc1a1f83a
            9d329777a7084e4aadae7122ce97005930528eb3c7f7f1129b372887a371155a
            3ba201a25cbf1dcb64e7cdee092c3141fb5550fe3d0dd82e870e578b2b465008
            18113b8f6569773c677385b69a42b77dcba7acffd95fd4452e23aaa1d37e1da2
            151ea658d40a3596b27ac9f8129dc6cf0643772624b59f4f461230df471ca260
            87c3942d5c6687df6082835935a3f87cb762b0c3b1d0dda4a6533965bef1b7b8
            292e254c014d090fed857c44c1839c694c0a64e3fad90a11f534722b6ee1574f
            2e149d55d744de4887024e08511431c062750e16c74ab9f3242f2db3ffb12a8d
            6107faa229d6f6373b07f36d3932b3bdb04c19dd64eadd7f93c3c564c358a1c8
            1dcf1c9c31e5b06568f97544c17dc15698c5cb38983a9afc42783faa773a52c9
            d8260690be9e3156aa5bc1509dea3f69587695cd6ff172ba83e6a6d8a7d6bbeb
            bbcda3672731983f89bc5831dc37c3f3c5c56facc697f3cb20bd5dbadbd702e5
            4844ac2f626901fe159db93dfd4773d8fe73562b846c1fc856d1802762840ebc
            72d7988bde75cbca70d319d32ce0cc0253bb2ad455723ee0c7f4736ce6e6665c
            5aca32a481c53839bc259167b013d0423395eeb9aaaee3206149a7d550d67fc5
            fdfe4a8a5c35d2510b664379ab8f72855a2af47abce2a632048eaf89e5cb4a88
            debc53a595103acce4f1cff18acff07afe1eb5716aa1e40b63134c3a3ae9579f
            a87f515be093c2d29db6d6b65c93661e00636b592704d093cc6716c2342eb185
            3d48c85c63ac8a2854462c7b77e7e3bd1eac5bca28ffaa00b5d349f8a547ad87
            5b96a8c2b2910c9301309a3f9138a5693111f55b3c009ca947c39dfc82d98eb1
            caa4a9cbe885f786fa86e55be062222f8ba90a974073326b31212aece0a34a60
            """
        )

        self.assertEqual(len(pk), len(exp_pk))
        self.assertEqual(pk, exp_pk)


class TestML_DSA_KAT(unittest.TestCase):
    """
    Test ML-DSA against test vectors collected from
    https://github.com/usnistgov/ACVP-Server/releases/tag/v1.1.0.35
    """

    def generic_keygen_kat(self, ML_DSA, index):
        with open("assets/ML-DSA-keyGen-FIPS204/internalProjection.json") as f:
            data = json.load(f)
        kat_data = data["testGroups"][index]["tests"]

        for test in kat_data:
            seed = bytes.fromhex(test["seed"])
            pk_kat = bytes.fromhex(test["pk"])
            sk_kat = bytes.fromhex(test["sk"])

            pk, sk = ML_DSA._keygen_internal(seed)
            self.assertEqual(pk, pk_kat)
            self.assertEqual(sk, sk_kat)

    def generic_sign_kat(self, ML_DSA, index, deterministic=False):
        with open("assets/ML-DSA-sigGen-FIPS204/internalProjection.json") as f:
            data = json.load(f)
        if deterministic:
            kat_data = data["testGroups"][2 * index]["tests"]
        else:
            kat_data = data["testGroups"][2 * index + 1]["tests"]

        for test in kat_data:
            sk_kat = bytes.fromhex(test["sk"])
            msg_kat = bytes.fromhex(test["message"])
            sig_kat = bytes.fromhex(test["signature"])

            if deterministic:
                rng_kat = bytes([0]) * 32
            else:
                rng_kat = bytes.fromhex(test["rng"])

            sig = ML_DSA._sign_internal(sk_kat, msg_kat, rng_kat)
            self.assertEqual(sig, sig_kat)

    def generic_verify_kat(self, ML_DSA, index):
        with open("assets/ML-DSA-sigVer-FIPS204/internalProjection.json") as f:
            data = json.load(f)
        pk_kat = bytes.fromhex(data["testGroups"][index]["pk"])
        kat_data = data["testGroups"][index]["tests"]

        for test in kat_data:
            check_kat = test["testPassed"]
            msg_kat = bytes.fromhex(test["message"])
            sig_kat = bytes.fromhex(test["signature"])

            check = ML_DSA._verify_internal(pk_kat, msg_kat, sig_kat)
            self.assertEqual(check, check_kat)

    def test_ML_DSA_44_keygen(self):
        self.generic_keygen_kat(ML_DSA_44, 0)

    def test_ML_DSA_65_keygen(self):
        self.generic_keygen_kat(ML_DSA_65, 1)

    def test_ML_DSA_87_keygen(self):
        self.generic_keygen_kat(ML_DSA_87, 2)

    def test_ML_DSA_44_sign(self):
        self.generic_sign_kat(ML_DSA_44, 0, deterministic=True)

    def test_ML_DSA_65_sign(self):
        self.generic_sign_kat(ML_DSA_65, 1, deterministic=True)

    def test_ML_DSA_87_sign(self):
        self.generic_sign_kat(ML_DSA_87, 2, deterministic=True)

    def test_ML_DSA_44_verify(self):
        self.generic_verify_kat(ML_DSA_44, 0)

    def test_ML_DSA_65_verify(self):
        self.generic_verify_kat(ML_DSA_65, 1)

    def test_ML_DSA_87_verify(self):
        self.generic_verify_kat(ML_DSA_87, 2)



class MatrixGeneric:
    def __init__(self, parent, matrix_data, transpose=False):
        self.parent = parent
        self._data = matrix_data
        self._transpose = transpose
        if not self._check_dimensions():
            raise ValueError("Inconsistent row lengths in matrix")

    def _check_dimensions(self):
        """
        Ensure that the matrix is rectangular
        """
        return len(set(map(len, self._data))) == 1

    def dim(self):
        """
        Return the dimensions of the matrix with m rows
        and n columns

        :return: the dimension of the matrix ``(m, n)``
        :rtype: tuple(int, int)
        """
        if not self._transpose:
            return len(self._data), len(self._data[0])
        else:
            return len(self._data[0]), len(self._data)

    def transpose(self):
        """
        Return a matrix with the rows and columns of swapped
        """
        return self.parent(self._data, not self._transpose)

    def transpose_self(self):
        """
        Swap the rows and columns of the matrix in place
        """
        self._transpose = not self._transpose
        return

    T = property(transpose)

    def reduce_coefficients(self):
        """
        Reduce every element in the polynomial
        using the modulus of the PolynomialRing
        """
        for row in self._data:
            for ele in row:
                ele.reduce_coefficients()
        return self

    def __getitem__(self, idx):
        """
        matrix[i, j] returns the element on row i, column j
        """
        assert isinstance(idx, tuple) and len(idx) == 2, "Can't access individual rows"
        if not self._transpose:
            return self._data[idx[0]][idx[1]]
        else:
            return self._data[idx[1]][idx[0]]

    def rows(self):
        """
        Return the rows of `self`
        """
        if not self._transpose:
            return zip(*self._data)
        else:
            return self._data

    def __eq__(self, other):
        if self.dim() != other.dim():
            return False
        m, n = self.dim()
        return all([self[i, j] == other[i, j] for i in range(m) for j in range(n)])

    def __neg__(self):
        """
        Returns -self, by negating all elements
        """
        m, n = self.dim()
        return self.parent(
            [[-self[i, j] for j in range(n)] for i in range(m)],
            False,
        )

    def __add__(self, other):
        if not isinstance(other, MatrixGeneric):
            raise TypeError("Can only add matrices to other matrices")
        if self.parent != other.parent:
            raise TypeError("Matrices must have the same base ring")
        if self.dim() != other.dim():
            raise ValueError("Matrices are not of the same dimensions")

        m, n = self.dim()
        return self.parent(
            [[self[i, j] + other[i, j] for j in range(n)] for i in range(m)],
            False,
        )

    def __iadd__(self, other):
        self = self + other
        return self

    def __sub__(self, other):
        if not isinstance(other, MatrixGeneric):
            raise TypeError("Can only add matrices to other matrices")
        if self.parent != other.parent:
            raise TypeError("Matrices must have the same base ring")
        if self.dim() != other.dim():
            raise ValueError("Matrices are not of the same dimensions")

        m, n = self.dim()
        return self.parent(
            [[self[i, j] - other[i, j] for j in range(n)] for i in range(m)],
            False,
        )

    def __isub__(self, other):
        self = self - other
        return self

    def __matmul__(self, other):
        """
        Denoted A @ B
        """
        if not isinstance(other, MatrixGeneric):
            raise TypeError("Can only multiply matrcies with other matrices")
        if self.parent != other.parent:
            raise TypeError("Matrices must have the same base ring")

        m, n = self.dim()
        n_, l = other.dim()
        if not n == n_:
            raise ValueError("Matrices are of incompatible dimensions")

        return self.parent(
            [
                [sum(self[i, k] * other[k, j] for k in range(n)) for j in range(l)]
                for i in range(m)
            ]
        )

    def scale(self, other):
        """
        Multiply each element of the matrix by a polynomial or integer
        """
        if not (isinstance(other, self.parent.ring.element) or isinstance(other, int)):
            raise TypeError("Can only multiply elements with polynomials or integers")

        matrix = [[other * ele for ele in row] for row in self._data]
        return self.parent(matrix, transpose=self._transpose)

    def dot(self, other):
        """
        Compute the inner product of two vectors
        """
        if not isinstance(other, MatrixGeneric):
            raise TypeError("Can only perform dot product with other matrices")
        res = self.T @ other
        assert res.dim() == (1, 1)
        return res[0, 0]

    def __repr__(self):
        m, n = self.dim()

        if m == 1 and n == 1:
            return str(self._data[0])

        max_col_width = [max(len(str(self[i, j])) for i in range(m)) for j in range(n)]
        info = "]\n[".join(
            [
                ", ".join([f"{str(self[i, j]):>{max_col_width[j]}}" for j in range(n)])
                for i in range(m)
            ]
        )
        return f"[{info}]"


class ModuleGeneric:
    def __init__(self, ring):
        """
        Initialise a module over the ring ``ring``.
        """
        self.ring = ring
        self.matrix_element = MatrixGeneric

    def random_element(self, m, n):
        """
        Generate a random element of the module of dimension m x n

        :param int m: the number of rows in the matrix
        :param int m: the number of columns in tge matrix
        :return: an element of the module with dimension `m times n`
        """
        elements = [[self.ring.random_element() for _ in range(n)] for _ in range(m)]
        return self(elements)

    def __repr__(self):
        return f"Module over the commutative ring: {self.ring}"

    def __str__(self):
        return f"Module over the commutative ring: {self.ring}"

    def __call__(self, matrix_elements, transpose=False):
        if not isinstance(matrix_elements, list):
            raise TypeError(
                "elements of a module are matrices, built from elements of the base ring"
            )

        if isinstance(matrix_elements[0], list):
            for element_list in matrix_elements:
                if not all(isinstance(aij, self.ring.element) for aij in element_list):
                    raise TypeError(
                        f"All elements of the matrix must be elements of the ring: {self.ring}"
                    )
            return self.matrix_element(self, matrix_elements, transpose=transpose)

        elif isinstance(matrix_elements[0], self.ring.element):
            if not all(isinstance(aij, self.ring.element) for aij in matrix_elements):
                raise TypeError(
                    f"All elements of the matrix must be elements of the ring: {self.ring}"
                )
            return self.matrix_element(self, [matrix_elements], transpose=transpose)

        else:
            raise TypeError(
                "elements of a module are matrices, built from elements of the base ring"
            )



import unittest
import os
from dilithium_py.drbg.aes256_ctr_drbg import AES256_CTR_DRBG


class TestDRBG(unittest.TestCase):
    """
    Some small tests, as the general check for the DRBG is that
    the KAT vectors match with the assets within the mlkem and
    kyber tests.
    """

    def test_no_seed(self):
        # If seed is none, os.urandom is used instead
        seed = None
        drbg = AES256_CTR_DRBG(seed)
        self.assertNotEqual(drbg.entropy_input, None)

    def test_bad_seed(self):
        # if the seed length is not 48, the code fails
        seed = b"1"
        self.assertRaises(ValueError, lambda: AES256_CTR_DRBG(seed))
        seed = b"1" * 49
        self.assertRaises(ValueError, lambda: AES256_CTR_DRBG(seed))

    def test_personalization(self):
        seed = os.urandom(48)
        personalization = os.urandom(24)
        drbg = AES256_CTR_DRBG(seed, personalization)
        self.assertEqual(AES256_CTR_DRBG, type(drbg))
        self.assertEqual(32, len(drbg.random_bytes(32)))
        self.assertEqual(bytes, type(drbg.random_bytes(32)))

    def test_bad_personalization(self):
        # if the personalization is longer than 48 bytes, fail
        seed = os.urandom(48)
        personalization = os.urandom(49)
        self.assertRaises(ValueError, lambda: AES256_CTR_DRBG(seed, personalization))

    def test_additional(self):
        drbg = AES256_CTR_DRBG()
        additional = os.urandom(24)
        self.assertEqual(32, len(drbg.random_bytes(32, additional)))
        self.assertEqual(bytes, type(drbg.random_bytes(32, additional)))

    def test_bad_additional(self):
        # if the additional data is longer than 48 bytes, fail
        drbg = AES256_CTR_DRBG()
        additional = os.urandom(49)
        self.assertRaises(ValueError, lambda: drbg.random_bytes(32, additional))



import random
import timeit

random.seed(0)
coeffs = [random.randint(0, 2**10) for _ in range(256)]

"""
Implementation of bitpacking of various polynomials
following section 5.2 of the Dilithium specification

https://pq-crystals.org/dilithium/data/dilithium-specification-round3-20210208.pdf
"""


def bit_pack_t1_old(coeffs):
    packed_bytes = []
    lim = 256 // 4
    for i in range(lim):
        packed_bytes.append((coeffs[4 * i + 0] >> 0) % 256)
        packed_bytes.append(((coeffs[4 * i + 0] >> 8) | (coeffs[4 * i + 1] << 2)) % 256)
        packed_bytes.append(((coeffs[4 * i + 1] >> 6) | (coeffs[4 * i + 2] << 4)) % 256)
        packed_bytes.append(((coeffs[4 * i + 2] >> 4) | (coeffs[4 * i + 3] << 6)) % 256)
        packed_bytes.append((coeffs[4 * i + 3] >> 2) % 256)
    return bytes(packed_bytes)


def bit_pack_1(coeffs):
    """
    Takes the coefficients of the polynomial and packs them
    as `n_bits` length bit strings
    """
    s = "".join(bin(c)[2:].zfill(10) for c in coeffs[::-1])
    return int(s, 2).to_bytes((len(s) + 7) // 8, byteorder="little")


def bit_pack_2(coeffs):
    r = 0
    for c in reversed(coeffs):
        r <<= 10
        r |= c
    return r.to_bytes(320, "little")


assert bit_pack_1(coeffs) == bit_pack_2(coeffs)


def bit_unpack_1(input_bytes):
    bytes_int = int.from_bytes(input_bytes, "little")
    s = bin(bytes_int)[2:].zfill(8 * len(input_bytes))
    coefficients = [int(s[i : i + 10], 2) for i in range(0, len(s), 10)][::-1]
    return coefficients


def bit_unpack_2(input_bytes):
    r = int.from_bytes(input_bytes, "little")
    mask = (1 << 10) - 1
    coefficients = []
    for _ in range(256):
        coefficients.append(r & mask)
        r >>= 10
    return coefficients


def bit_unpack_3(input_bytes):
    r = int.from_bytes(input_bytes, "little")
    mask = (1 << 10) - 1
    return [(r >> 10 * i) & mask for i in range(256)]


packed = bit_pack_2(coeffs)
assert coeffs == bit_unpack_1(packed)
assert coeffs == bit_unpack_2(packed)
assert coeffs == bit_unpack_3(packed)

print("Packing with binary strings")
print(timeit.timeit("bit_pack_1(coeffs)", globals=globals(), number=10_000))
print("Packing with bitshifts")
print(timeit.timeit("bit_pack_2(coeffs)", globals=globals(), number=10_000))

print("Unpacking with binary strings")
print(timeit.timeit("bit_unpack_1(packed)", globals=globals(), number=10_000))
print("Unpacking with bitshifts")
print(timeit.timeit("bit_unpack_2(packed)", globals=globals(), number=10_000))
print("Unpacking with bitshifts")
print(timeit.timeit("bit_unpack_3(packed)", globals=globals(), number=10_000))



import unittest
from dilithium_py.utilities.utils import reduce_mod_pm
from random import randint


class TestUtils(unittest.TestCase):
    def test_reduce_mod_pm_even(self):
        """
        For odd modulus the reduced value should be bounded by
        -(n-1)/2 <= x <= (n-1)/2
        """
        for _ in range(100):
            modulus = 2 * randint(0, 100)
            for i in range(modulus):
                x = reduce_mod_pm(i, modulus)
                self.assertTrue(x <= modulus // 2)
                self.assertTrue(-modulus // 2 < x)

    def test_reduce_mod_pm_odd(self):
        """
        For even modulus the reduced value should be bounded by
        -n/2 < x <= n/2
        """
        for _ in range(100):
            modulus = 2 * randint(0, 100) + 1
            for i in range(modulus):
                x = reduce_mod_pm(i, modulus)
                self.assertTrue(x <= (modulus - 1) // 2)
                self.assertTrue(-(modulus - 1) // 2 <= x)



from dilithium_py.dilithium import Dilithium2, Dilithium3, Dilithium5
import cProfile
from time import time
from statistics import mean, median


def profile_dilithium(Dilithium):
    pk, sk = Dilithium.keygen()
    m = b"Signed by dilithium"
    sig = Dilithium.sign(sk, m)
    check = Dilithium.verify(pk, m, sig)
    assert check

    gvars = {}
    lvars = {"Dilithium": Dilithium, "m": m, "pk": pk, "sk": sk, "sig": sig}

    cProfile.runctx(
        "[Dilithium.keygen() for _ in range(100)]", globals=gvars, locals=lvars, sort=1
    )
    cProfile.runctx(
        "[Dilithium.sign(sk, m) for _ in range(100)]",
        globals=gvars,
        locals=lvars,
        sort=1,
    )
    cProfile.runctx(
        "[Dilithium.verify(pk, m, sig) for _ in range(100)]",
        globals=gvars,
        locals=lvars,
        sort=1,
    )


def benchmark_dilithium(Dilithium, name, count):
    # Banner
    print("-" * 27)
    print(f"  {name} | ({count} calls)")
    print("-" * 27)

    fails = 0
    keygen_times = []
    sign_times = []
    verify_times = []
    # 32 byte message
    m = b"Your message signed by Dilithium"

    for _ in range(count):
        t0 = time()
        pk, sk = Dilithium.keygen()
        keygen_times.append(time() - t0)

        t1 = time()
        sig = Dilithium.sign(sk, m)
        sign_times.append(time() - t1)

        t2 = time()
        verify = Dilithium.verify(pk, m, sig)
        verify_times.append(time() - t2)
        if not verify:
            fails += 1

    print(f"Keygen median: {round(median(keygen_times), 3)}")
    print(f"Sign median: {round(median(sign_times), 3)}")
    print(f"Sign average: {round(mean(sign_times), 3)}")
    print(f"Verify median: {round(median(verify_times), 3)}")
    print(f"Fails: {fails}")


if __name__ == "__main__":
    # I used 1000 calls for the README, but you might want to
    # shrink this down if you're experimenting
    count = 1000
    benchmark_dilithium(Dilithium2, "Dilithium2", count)
    benchmark_dilithium(Dilithium3, "Dilithium3", count)
    benchmark_dilithium(Dilithium5, "Dilithium5", count)



from hashlib import shake_128, shake_256
from dilithium_py.shake.shake_wrapper import shake128, shake256
from Crypto.Hash.SHAKE128 import SHAKE128_XOF
from Crypto.Hash.SHAKE256 import SHAKE256_XOF

import unittest
import random


class TestShakeHashlib(unittest.TestCase):
    def hashlib_test_long_calls(self, Shake, shake_hashlib):
        absorb_bytes = b"testing_shake_long"
        for l in [1, 100, 1000, 2000, 5000, 1_000_000]:
            Shake.absorb(absorb_bytes)
            self.assertEqual(shake_hashlib(absorb_bytes).digest(l), Shake.read(l))

    def hashlib_test_many_calls(self, Shake, shake_hashlib):
        absorb_bytes = b"testing_shake_one"
        for l in [1, 100, 1000, 2000, 5000, 1_000_000]:
            Shake.absorb(absorb_bytes)
            output = b"".join([Shake.read(1) for _ in range(l)])
            self.assertEqual(shake_hashlib(absorb_bytes).digest(l), output)

    def test_hashlib_shake128(self):
        self.hashlib_test_long_calls(shake128, shake_128)
        self.hashlib_test_many_calls(shake128, shake_128)

    def test_hashlib_shake256(self):
        self.hashlib_test_long_calls(shake256, shake_256)
        self.hashlib_test_many_calls(shake256, shake_256)


class TestShakeCrypto(unittest.TestCase):
    def pycryptodome_test_read_chunks(self, Shake, ShakeCrypto):
        absorb_bytes = b"testing_shake_chunks"
        chunk = random.randint(50, 100)
        Shake.absorb(absorb_bytes)
        ShakeCrypto.update(absorb_bytes)
        for _ in range(1000):
            self.assertEqual(Shake.read(chunk), ShakeCrypto.read(chunk))

    def test_pycryptodome_shake(self):
        self.pycryptodome_test_read_chunks(shake128, SHAKE128_XOF())
        self.pycryptodome_test_read_chunks(shake256, SHAKE256_XOF())



from .default_parameters import (
    ML_DSA_44,
    ML_DSA_65,
    ML_DSA_87,
    HASH_ML_DSA_44_WITH_SHA512,
    HASH_ML_DSA_65_WITH_SHA512,
    HASH_ML_DSA_87_WITH_SHA512,
)

__all__ = [
    "ML_DSA_44",
    "ML_DSA_65",
    "ML_DSA_87",
    "HASH_ML_DSA_44_WITH_SHA512",
    "HASH_ML_DSA_65_WITH_SHA512",
    "HASH_ML_DSA_87_WITH_SHA512",
]



from hashlib import shake_128, shake_256

"""
hashlib has implemented shake_128, shake_256 but they haven't designed it so you
can read bytes properly... every call generates all bytes without updating

shake_128.digest(1) == shake_128.digest(1)

and we have no shake_128.read() :(

So, here's a wrapper which calls to shake_128.digest and collects a bunch of
bytes which we can then read through.
"""


class Shake:
    def __init__(self, algorithm, block_length):
        self.algorithm = algorithm
        self.block_length = block_length
        self.buf = b""
        self.len_buf = 0

    def absorb(self, input_bytes):
        """
        Initialise the XOF with the seed and reset other init.
        """
        # Initalize the buffer
        self.index = 0

        # Set the reading method from hashlib digest
        self.xof_read = self.algorithm(input_bytes).digest

        # Start by requesting 5 blocks from the XOF
        self.buf = self.xof_read(5 * self.block_length)
        self.len_buf = 5 * self.block_length

    def read(self, n):
        """
        Read n bytes from the XOF
        """
        # Make sure there are enough bytes to read
        while self.index + n > self.len_buf:
            # double the size of the buffer
            self.len_buf *= 2
            self.buf = self.xof_read(self.len_buf)

        # Read from the buffer data the bytes requested
        send = self.buf[self.index : self.index + n]

        # Shift the index along the buffer
        self.index += n

        return send

    def __call__(self, input_bytes):
        self.absorb(input_bytes)
        return self


shake128 = Shake(shake_128, 168)
shake256 = Shake(shake_256, 136)



import unittest
from random import randint
from dilithium_py.polynomials.polynomials_generic import PolynomialRingGeneric
from dilithium_py.modules.modules_generic import ModuleGeneric


class TestModule(unittest.TestCase):
    R = PolynomialRingGeneric(11, 5)
    M = ModuleGeneric(R)

    def test_random_element(self):
        for _ in range(100):
            m = randint(1, 5)
            n = randint(1, 5)
            A = self.M.random_element(m, n)
            self.assertEqual(type(A), self.M.matrix_element)
            self.assertEqual(type(A[0, 0]), self.R.element)
            self.assertEqual(A.dim(), (m, n))

    def test_print(self):
        s = "Module over the commutative ring: Univariate Polynomial Ring in x over Finite Field of size 11 with modulus x^5 + 1"
        self.assertEqual(str(self.M), s)
        self.assertEqual(self.M.__repr__(), s)

    def test_non_list_error(self):
        self.assertRaises(TypeError, lambda: self.M("1"))

    def test_non_ring_list_error(self):
        one = self.R(1)
        self.assertRaises(TypeError, lambda: self.M([one, "2", "3"]))
        self.assertRaises(TypeError, lambda: self.M(["2", one, "3"]))
        self.assertRaises(TypeError, lambda: self.M([[one, "2", "3"], [one, "2", "3"]]))
        self.assertRaises(TypeError, lambda: self.M([["1", one, "3"], [one, "2", "3"]]))

    def test_non_rectangular(self):
        one = self.R(1)
        self.assertRaises(ValueError, lambda: self.M([[one, one], [one]]))


class TestMatrix(unittest.TestCase):
    R = PolynomialRingGeneric(11, 5)
    R_prime = PolynomialRingGeneric(11, 2)
    M = ModuleGeneric(R)
    M_prime = ModuleGeneric(R)

    def test_equality(self):
        for _ in range(100):
            A = self.M.random_element(2, 2)
            B = self.M.random_element(2, 3)

            self.assertEqual(A, A)
            self.assertNotEqual(A, B)

    def test_add_errors(self):
        A = self.M.random_element(2, 2)
        B = self.M.random_element(2, 3)
        A_prime = self.M_prime.random_element(2, 2)

        self.assertRaises(TypeError, lambda: A + "B")
        self.assertRaises(ValueError, lambda: A + B)
        self.assertRaises(TypeError, lambda: A + A_prime)

    def test_matrix_add(self):
        zero = self.R(0)
        Z = self.M([[zero, zero], [zero, zero]])
        for _ in range(100):
            A = self.M.random_element(2, 2)
            B = self.M.random_element(2, 2)
            C = self.M.random_element(2, 2)

            self.assertEqual(A + Z, A)
            self.assertEqual(A + B, B + A)
            self.assertEqual(A + (B + C), (A + B) + C)

            B = C
            B += C
            self.assertEqual(B, C + C)

    def test_sub_errors(self):
        A = self.M.random_element(2, 2)
        B = self.M.random_element(2, 3)
        A_prime = self.M_prime.random_element(2, 2)

        self.assertRaises(TypeError, lambda: A - "B")
        self.assertRaises(ValueError, lambda: A - B)
        self.assertRaises(TypeError, lambda: A - A_prime)

    def test_matrix_sub(self):
        zero = self.R(0)
        Z = self.M([[zero, zero], [zero, zero]])
        for _ in range(100):
            A = self.M.random_element(2, 2)
            B = self.M.random_element(2, 2)
            C = self.M.random_element(2, 2)

            self.assertEqual(A - Z, A)
            self.assertEqual(A - B, -(B - A))
            self.assertEqual(A - (B - C), (A - B) + C)

            B = C
            B -= C
            self.assertEqual(B, Z)

    def test_mul_errors(self):
        A = self.M.random_element(2, 2)
        B = self.M.random_element(5, 5)
        A_prime = self.M_prime.random_element(2, 2)

        self.assertRaises(TypeError, lambda: A @ "B")
        self.assertRaises(ValueError, lambda: A @ B)
        self.assertRaises(TypeError, lambda: A @ A_prime)

    def test_matrix_mul_square(self):
        zero = self.R(0)
        one = self.R(1)
        Z = self.M([[zero, zero], [zero, zero]])
        I = self.M([[one, zero], [zero, one]])
        for _ in range(100):
            A = self.M.random_element(2, 2)
            B = self.M.random_element(2, 2)
            C = self.M.random_element(2, 2)
            d = self.R.random_element()
            D = self.M([[d, zero], [zero, d]])

            self.assertEqual(A @ Z, Z)
            self.assertEqual(A @ I, A)
            self.assertEqual(A @ D, D @ A)  # Diagonal matrices commute
            self.assertEqual(A @ (B + C), A @ B + A @ C)

    def test_matrix_mul_rectangle(self):
        for _ in range(100):
            A = self.M.random_element(7, 3)
            B = self.M.random_element(3, 2)
            C = self.M.random_element(3, 2)

            self.assertEqual(A @ (B + C), A @ B + A @ C)

    def test_matrix_transpose_id(self):
        zero = self.R(0)
        one = self.R(1)
        I = self.M([[one, zero], [zero, one]])

        self.assertEqual(I, I.transpose())

    def test_matrix_transpose(self):
        for _ in range(100):
            A = self.M.random_element(7, 3)
            At = A.transpose()
            AAt = A @ At

            # Should always be symmetric
            self.assertEqual(AAt, AAt.transpose())

            # Assert transpose in place works
            At.transpose_self()
            self.assertEqual(A, At)

    def test_matrix_dot(self):
        for _ in range(100):
            u = [self.R.random_element() for _ in range(5)]
            v = [self.R.random_element() for _ in range(5)]
            dot = sum([ui * vi for ui, vi in zip(u, v)])

            U = self.M([u], transpose=True)
            V = self.M([v], transpose=True)

            self.assertEqual(dot, U.dot(V))

        u = [self.R.random_element() for _ in range(5)]
        U = self.M([u])
        self.assertRaises(TypeError, lambda: U.dot("A"))

    def test_print(self):
        A = self.M(
            [
                [self.R([1, 2]), self.R([3, 4, 5, 6])],
                [self.R([0, 0, 0, 0, 3]), self.R([0, 1, 0, 3])],
            ],
        )
        u = self.M([self.R([1, 2]), self.R([3, 4, 5, 6])])

        sA = "[1 + 2*x, 3 + 4*x + 5*x^2 + 6*x^3]\n[  3*x^4,               x + 3*x^3]"
        su = "[1 + 2*x, 3 + 4*x + 5*x^2 + 6*x^3]"
        self.assertEqual(str(A), sA)
        self.assertEqual(str(u), su)

        A = self.M([self.R(1), self.R(2), self.R(3)])
        sA = "[1, 2, 3]"
        self.assertEqual(str(A), sA)

        A = self.M([[self.R(1)], [self.R(2)], [self.R(3)]])
        sA = "[1]\n[2]\n[3]"
        self.assertEqual(str(A), sA)



from typing import cast
from .modules_generic import ModuleGeneric, MatrixGeneric
from ..polynomials.polynomials import PolynomialRing


class Matrix(MatrixGeneric):
    def __init__(self, parent, matrix_data, transpose=False):
        super().__init__(parent, matrix_data, transpose=transpose)

    def check_norm_bound(self, bound):
        for row in self._data:
            if any(p.check_norm_bound(bound) for p in row):
                return True
        return False

    def power_2_round(self, d):
        """
        Applies `power_2_round` on every element in the
        Matrix to create two matrices.
        """
        m, n = self.dim()

        m1_elements = [[0 for _ in range(n)] for _ in range(m)]
        m0_elements = [[0 for _ in range(n)] for _ in range(m)]

        for i in range(m):
            for j in range(n):
                m1_ele, m0_ele = self[i, j].power_2_round(d)
                m1_elements[i][j] = m1_ele
                m0_elements[i][j] = m0_ele

        return self.parent(m1_elements, transpose=self._transpose), self.parent(
            m0_elements, transpose=self._transpose
        )

    def decompose(self, alpha):
        """
        Applies `power_2_round` on every element in the
        Matrix to create two matrices.
        """
        m, n = self.dim()

        m1_elements = [[0 for _ in range(n)] for _ in range(m)]
        m0_elements = [[0 for _ in range(n)] for _ in range(m)]

        for i in range(m):
            for j in range(n):
                m1_ele, m0_ele = self[i, j].decompose(alpha)
                m1_elements[i][j] = m1_ele
                m0_elements[i][j] = m0_ele

        return self.parent(m1_elements, transpose=self._transpose), self.parent(
            m0_elements, transpose=self._transpose
        )

    def __bit_pack(self, algorithm, *args):
        return b"".join(algorithm(poly, *args) for row in self.rows() for poly in row)

    def bit_pack_t1(self):
        algorithm = self.parent.ring.element.bit_pack_t1
        return self.__bit_pack(algorithm)

    def bit_pack_t0(self):
        algorithm = self.parent.ring.element.bit_pack_t0
        return self.__bit_pack(algorithm)

    def bit_pack_s(self, eta):
        algorithm = self.parent.ring.element.bit_pack_s
        return self.__bit_pack(algorithm, eta)

    def bit_pack_w(self, gamma_2):
        algorithm = self.parent.ring.element.bit_pack_w
        return self.__bit_pack(algorithm, gamma_2)

    def bit_pack_z(self, gamma_1):
        algorithm = self.parent.ring.element.bit_pack_z
        return self.__bit_pack(algorithm, gamma_1)

    def to_ntt(self):
        """
        Convert every element of the matrix into NTT form
        """
        data = [[x.to_ntt() for x in row] for row in self._data]
        return self.parent(data, self._transpose)

    def from_ntt(self):
        """
        Convert every element of the matrix from NTT form
        """
        data = [[x.from_ntt() for x in row] for row in self._data]
        return self.parent(data, self._transpose)

    def high_bits(self, alpha, is_ntt=False):
        matrix = [
            [ele.high_bits(alpha, is_ntt=is_ntt) for ele in row] for row in self.rows()
        ]
        return self.parent(matrix)

    def low_bits(self, alpha, is_ntt=False):
        matrix = [
            [ele.low_bits(alpha, is_ntt=is_ntt) for ele in row] for row in self.rows()
        ]
        return self.parent(matrix)

    def make_hint(self, other, alpha):
        """
        Figure 3 (Supporting algorithms for Dilithium)
        https://pq-crystals.org/dilithium/data/dilithium-specification-round3-20210208.pdf
        """
        matrix = [
            [p.make_hint(q, alpha) for p, q in zip(r1, r2)]
            for r1, r2 in zip(self.rows(), other.rows())
        ]
        return self.parent(matrix)

    def make_hint_optimised(self, other, alpha):
        """
        Figure 3 (Supporting algorithms for Dilithium)
        https://pq-crystals.org/dilithium/data/dilithium-specification-round3-20210208.pdf
        """
        matrix = [
            [p.make_hint_optimised(q, alpha) for p, q in zip(r1, r2)]
            for r1, r2 in zip(self.rows(), other.rows())
        ]
        return self.parent(matrix)

    def use_hint(self, other, alpha):
        """
        Figure 3 (Supporting algorithms for Dilithium)
        https://pq-crystals.org/dilithium/data/dilithium-specification-round3-20210208.pdf
        """
        matrix = [
            [p.use_hint(q, alpha) for p, q in zip(r1, r2)]
            for r1, r2 in zip(self.rows(), other.rows())
        ]
        return self.parent(matrix)

    def sum_hint(self):
        """
        Helper function to count the number of coeffs == 1
        in all the polynomials of a matrix
        """
        return sum(c for row in self.rows() for p in row for c in p)


class Vector(Matrix):
    def __init__(self, parent, vector_elements):
        super().__init__(parent, [vector_elements], transpose=True)


class Module(ModuleGeneric):
    def __init__(self):
        self.ring = PolynomialRing()
        self.matrix_element = Matrix
        self.vector_element = Vector

    def __call__(self, matrix_elements, transpose=False) -> Matrix:
        """
        NOTE: This is simply a wrapper function of the Module method to ensure correct
        type casting
        """
        mat = super().__call__(matrix_elements, transpose)
        return cast(Matrix, mat)

    def vector(self, elements) -> Vector:
        """
        Construct a vector given a list of elements of the module's ring

        :param list: a list of elements of the ring
        :return: a vector of the module
        """
        return Vector(self, elements)

    def __bit_unpack(self, input_bytes, k, alg, packed_len, *args) -> Vector:
        poly_bytes = [
            input_bytes[i : i + packed_len]
            for i in range(0, len(input_bytes), packed_len)
        ]
        matrix = [alg(poly_bytes[i], *args) for i in range(k)]
        return self.vector(matrix)

    def bit_unpack_t0(self, input_bytes, k) -> Vector:
        packed_len = 416
        algorithm = self.ring.bit_unpack_t0
        return self.__bit_unpack(input_bytes, k, algorithm, packed_len)

    def bit_unpack_t1(self, input_bytes, k):
        packed_len = 320
        algorithm = self.ring.bit_unpack_t1
        return self.__bit_unpack(input_bytes, k, algorithm, packed_len)

    def bit_unpack_s(self, input_bytes, k, eta) -> Vector:
        # Level 2 and 5 parameter set
        if eta == 2:
            packed_len = 96
        # Level 3 parameter set
        elif eta == 4:
            packed_len = 128
        else:
            raise ValueError("Expected eta to be either 2 or 4")
        algorithm = self.ring.bit_unpack_s
        return self.__bit_unpack(input_bytes, k, algorithm, packed_len, eta)

    def bit_unpack_w(self, input_bytes, k, gamma_2) -> Vector:
        # Level 2 parameter set
        if gamma_2 == 95232:
            packed_len = 192
        # Level 3 and 5 parameter set
        elif gamma_2 == 261888:
            packed_len = 128
        else:
            raise ValueError("Expected gamma_2 to be either (q-1)/88 or (q-1)/32")
        algorithm = self.ring.bit_unpack_w
        return self.__bit_unpack(input_bytes, k, algorithm, packed_len, gamma_2)

    def bit_unpack_z(self, input_bytes, k, gamma_1) -> Vector:
        # Level 2 parameter set
        if gamma_1 == (1 << 17):
            packed_len = 576
        # Level 3 and 5 parameter set
        elif gamma_1 == (1 << 19):
            packed_len = 640
        else:
            raise ValueError("Expected gamma_1 to be either 2^17 or 2^19")
        algorithm = self.ring.bit_unpack_z
        return self.__bit_unpack(input_bytes, k, algorithm, packed_len, gamma_1)



import os
from ..modules.modules import Module

try:
    from xoflib import shake256
except ImportError:
    from ..shake.shake_wrapper import shake256


class Dilithium:
    def __init__(self, parameter_set):
        self.d = parameter_set["d"]
        self.k = parameter_set["k"]
        self.l = parameter_set["l"]
        self.eta = parameter_set["eta"]
        self.tau = parameter_set["tau"]
        self.omega = parameter_set["omega"]
        self.gamma_1 = parameter_set["gamma_1"]
        self.gamma_2 = parameter_set["gamma_2"]
        self.beta = self.tau * self.eta

        self.M = Module()
        self.R = self.M.ring

        # Use system randomness by default, for deterministic randomness
        # use the method `set_drbg_seed()`
        self.random_bytes = os.urandom

    def set_drbg_seed(self, seed):
        """
        Change entropy source to a DRBG and seed it with provided value.

        Setting the seed switches the entropy source from :func:`os.urandom()`
        to an AES256 CTR DRBG.

        Used for both deterministic versions of Kyber as well as testing
        alignment with the KAT vectors

        Note:
          currently requires pycryptodome for AES impl.
        """
        try:
            from ..drbg.aes256_ctr_drbg import AES256_CTR_DRBG

            self._drbg = AES256_CTR_DRBG(seed)
            self.random_bytes = self._drbg.random_bytes
        except ImportError as e:  # pragma: no cover
            print(f"Error importing AES from pycryptodome: {e = }")
            raise Warning(
                "Cannot set DRBG seed due to missing dependencies, try installing requirements: pip -r install requirements"
            )

    """
    H() uses Shake256 to hash data to 32 and 64 bytes in a
    few places in the code
    """

    @staticmethod
    def _h(input_bytes, length):
        """
        H: B^*  -> B^*
        """
        return shake256(input_bytes).read(length)

    def _expand_matrix_from_seed(self, rho):
        """
        Helper function which generates a element of size
        k x l from a seed `rho`.
        """
        A_data = [[self.R.zero() for _ in range(self.l)] for _ in range(self.k)]
        for i in range(self.k):
            for j in range(self.l):
                A_data[i][j] = self.R.rejection_sample_ntt_poly(rho, i, j)
        return self.M(A_data)

    def _expand_vector_from_seed(self, rho_prime):
        s1_elements = [
            self.R.rejection_bounded_poly(rho_prime, i, self.eta) for i in range(self.l)
        ]
        s2_elements = [
            self.R.rejection_bounded_poly(rho_prime, i, self.eta)
            for i in range(self.l, self.l + self.k)
        ]

        s1 = self.M.vector(s1_elements)
        s2 = self.M.vector(s2_elements)
        return s1, s2

    def _expand_mask_vector(self, rho_prime, kappa):
        elements = [
            self.R.sample_mask_polynomial(rho_prime, i, kappa, self.gamma_1)
            for i in range(self.l)
        ]
        return self.M.vector(elements)

    @staticmethod
    def _pack_pk(rho, t1):
        return rho + t1.bit_pack_t1()

    def _pack_sk(self, rho, K, tr, s1, s2, t0):
        s1_bytes = s1.bit_pack_s(self.eta)
        s2_bytes = s2.bit_pack_s(self.eta)
        t0_bytes = t0.bit_pack_t0()
        return rho + K + tr + s1_bytes + s2_bytes + t0_bytes

    def _pack_h(self, h):
        non_zero_positions = [
            [i for i, c in enumerate(poly.coeffs) if c == 1]
            for row in h._data
            for poly in row
        ]
        packed = []
        offsets = []
        for positions in non_zero_positions:
            packed.extend(positions)
            offsets.append(len(packed))

        padding_len = self.omega - offsets[-1]
        packed.extend([0 for _ in range(padding_len)])
        return bytes(packed + offsets)

    def _pack_sig(self, c_tilde, z, h):
        return c_tilde + z.bit_pack_z(self.gamma_1) + self._pack_h(h)

    def _unpack_pk(self, pk_bytes):
        rho, t1_bytes = pk_bytes[:32], pk_bytes[32:]
        t1 = self.M.bit_unpack_t1(t1_bytes, self.k)
        return rho, t1

    def _unpack_sk(self, sk_bytes):
        if self.eta == 2:
            s_bytes = 96
        else:
            s_bytes = 128
        s1_len = s_bytes * self.l
        s2_len = s_bytes * self.k
        t0_len = 416 * self.k
        if len(sk_bytes) != 3 * 32 + s1_len + s2_len + t0_len:
            raise ValueError("SK packed bytes is of the wrong length")

        # Split bytes between seeds and vectors
        sk_seed_bytes, sk_vec_bytes = sk_bytes[:96], sk_bytes[96:]

        # Unpack seed bytes
        rho, K, tr = (
            sk_seed_bytes[:32],
            sk_seed_bytes[32:64],
            sk_seed_bytes[64:96],
        )

        # Unpack vector bytes
        s1_bytes = sk_vec_bytes[:s1_len]
        s2_bytes = sk_vec_bytes[s1_len : s1_len + s2_len]
        t0_bytes = sk_vec_bytes[-t0_len:]

        # Unpack bytes to vectors
        s1 = self.M.bit_unpack_s(s1_bytes, self.l, self.eta)
        s2 = self.M.bit_unpack_s(s2_bytes, self.k, self.eta)
        t0 = self.M.bit_unpack_t0(t0_bytes, self.k)

        return rho, K, tr, s1, s2, t0

    def _unpack_h(self, h_bytes):
        offsets = [0] + list(h_bytes[-self.k :])
        # check offsets are monotonic increasing
        if any(offsets[i] > offsets[i + 1] for i in range(len(offsets) - 1)):
            raise ValueError("Offsets in h_bytes are not monotonic increasing")
        # check offset[-1] is smaller than the length of h_bytes
        if offsets[-1] > self.omega:
            raise ValueError("Accumulate offset of hints exceeds omega")
        # check zero fields are all zeros
        if any(b != 0 for b in h_bytes[offsets[-1] : self.omega]):
            raise ValueError("Non-zero fields in h_bytes are not all zeros")

        non_zero_positions = [
            list(h_bytes[offsets[i] : offsets[i + 1]]) for i in range(self.k)
        ]

        matrix = []
        for poly_non_zero in non_zero_positions:
            coeffs = [0 for _ in range(256)]
            for i, non_zero in enumerate(poly_non_zero):
                if i > 0 and non_zero < poly_non_zero[i - 1]:
                    raise ValueError(
                        "Non-zero positions in h_bytes are not monotonic increasing"
                    )
            for non_zero in poly_non_zero:
                coeffs[non_zero] = 1
            matrix.append([self.R(coeffs)])
        return self.M(matrix)

    def _unpack_sig(self, sig_bytes):
        c_tilde = sig_bytes[:32]
        z_bytes = sig_bytes[32 : -(self.k + self.omega)]
        h_bytes = sig_bytes[-(self.k + self.omega) :]

        z = self.M.bit_unpack_z(z_bytes, self.l, self.gamma_1)
        h = self._unpack_h(h_bytes)
        return c_tilde, z, h

    def keygen(self):
        """
        Generates a public-private keyair
        """
        # Random seed
        zeta = self.random_bytes(32)

        # Expand with an XOF (SHAKE256)
        seed_bytes = self._h(zeta, 128)

        # Split bytes into suitable chunks
        rho, rho_prime, K = seed_bytes[:32], seed_bytes[32:96], seed_bytes[96:]

        # Generate matrix A ∈ R^(kxl) in the NTT domain
        A_hat = self._expand_matrix_from_seed(rho)

        # Generate the error vectors s1 ∈ R^l, s2 ∈ R^k
        s1, s2 = self._expand_vector_from_seed(rho_prime)
        s1_hat = s1.to_ntt()

        # Matrix multiplication
        t = (A_hat @ s1_hat).from_ntt() + s2

        t1, t0 = t.power_2_round(self.d)

        # Pack up the bytes
        pk = self._pack_pk(rho, t1)
        tr = self._h(pk, 32)

        sk = self._pack_sk(rho, K, tr, s1, s2, t0)
        return pk, sk

    def sign(self, sk_bytes, m):
        """
        Generates a signature for a message m from a byte-encoded private key
        """
        # unpack the secret key
        rho, K, tr, s1, s2, t0 = self._unpack_sk(sk_bytes)

        # Generate matrix A ∈ R^(kxl) in the NTT domain
        A_hat = self._expand_matrix_from_seed(rho)

        # Set seeds and nonce (kappa)
        mu = self._h(tr + m, 64)
        kappa = 0
        rho_prime = self._h(K + mu, 64)

        # Precompute NTT representation
        s1 = s1.to_ntt()
        s2 = s2.to_ntt()
        t0 = t0.to_ntt()

        alpha = self.gamma_2 << 1
        while True:
            y = self._expand_mask_vector(rho_prime, kappa)
            y_hat = y.to_ntt()

            # increment the nonce
            kappa += self.l

            w = (A_hat @ y_hat).from_ntt()

            # Extract out both the high and low bits
            w1, w0 = w.decompose(alpha)

            # Create challenge polynomial
            w1_bytes = w1.bit_pack_w(self.gamma_2)
            c_tilde = self._h(mu + w1_bytes, 32)
            c = self.R.sample_in_ball(c_tilde, self.tau)

            # Store c in NTT form
            c = c.to_ntt()

            z = y + (s1.scale(c)).from_ntt()
            if z.check_norm_bound(self.gamma_1 - self.beta):
                continue

            w0_minus_cs2 = w0 - s2.scale(c).from_ntt()
            if w0_minus_cs2.check_norm_bound(self.gamma_2 - self.beta):
                continue

            c_t0 = t0.scale(c).from_ntt()
            if c_t0.check_norm_bound(self.gamma_2):
                continue

            w0_minus_cs2_plus_ct0 = w0_minus_cs2 + c_t0

            h = w0_minus_cs2_plus_ct0.make_hint_optimised(w1, alpha)
            if h.sum_hint() > self.omega:
                continue

            return self._pack_sig(c_tilde, z, h)

    def verify(self, pk_bytes, m, sig_bytes):
        """
        Verifies a signature for a message m from a byte encoded public key and
        signature
        """
        rho, t1 = self._unpack_pk(pk_bytes)
        try:
            c_tilde, z, h = self._unpack_sig(sig_bytes)
        except ValueError:
            # verify failed if malformed input signature
            return False

        if h.sum_hint() > self.omega:
            return False

        if z.check_norm_bound(self.gamma_1 - self.beta):
            return False

        A_hat = self._expand_matrix_from_seed(rho)

        tr = self._h(pk_bytes, 32)
        mu = self._h(tr + m, 64)
        c = self.R.sample_in_ball(c_tilde, self.tau)

        # Convert to NTT for computation
        c = c.to_ntt()
        z = z.to_ntt()

        t1 = t1.scale(1 << self.d)
        t1 = t1.to_ntt()

        Az_minus_ct1 = (A_hat @ z) - t1.scale(c)
        Az_minus_ct1 = Az_minus_ct1.from_ntt()

        w_prime = h.use_hint(Az_minus_ct1, 2 * self.gamma_2)
        w_prime_bytes = w_prime.bit_pack_w(self.gamma_2)

        return c_tilde == self._h(mu + w_prime_bytes, 32)



"""
Import and export of keys to PKCS #8 and SPKI format.

This module allows for importing private keys (signing keys) from
PKCS #8 files, exporting them to PKCS #8 files, and both import and export
of the public keys (verification keys) from the Subject Public Key Info
format used in X.509 certificates and in bare public keys.
"""

try:
    from ecdsa import der
except ImportError:
    raise ImportError("PKCS functionality requires the ecdsa library")


from .default_parameters import ML_DSA_44, ML_DSA_65, ML_DSA_87


OIDS = {
    ML_DSA_44.oid: ML_DSA_44,
    ML_DSA_65.oid: ML_DSA_65,
    ML_DSA_87.oid: ML_DSA_87,
}


def pk_to_der(ml_dsa, pk):
    """
    Convert a verification key to a SPKI DER structure.

    :param ml_dsa: an ``ML_DSA`` object instance
    :param bytes pk: verification key matching the passed ML_DSA object
    :rtype: bytes
    """
    if not ml_dsa.oid:
        raise ValueError("Only ML_DSAs with specified OIDs can be encoded")

    if len(pk) != ml_dsa._pk_size():
        raise ValueError("Provided key size doesn't match the provided ML_DSA")

    enc = der.encode_sequence(
        der.encode_sequence(
            der.encode_oid(*ml_dsa.oid),
        ),
        der.encode_bitstring(pk, 0),
    )

    return enc


def pk_to_pem(ml_dsa, pk):
    """
    Convert a verification key to a SPKI PEM structure.

    :param ml_dsa: an ``ML_DSA`` object instance
    :param bytes pk: verification key matching the passed ML_DSA object
    :rtype: str
    """
    der_enc = pk_to_der(ml_dsa, pk)

    pem_enc = der.topem(der_enc, "PUBLIC KEY")

    return pem_enc


def pk_from_der(enc_key):
    """
    Extract a verification key from a SPKI DER encoding.

    :param bytes enc_key: SPKI DER encoding of a key
    :rtype: tuple(ML_DSA, bytes)
    """
    s1, empty = der.remove_sequence(enc_key)
    if empty:
        raise der.UnexpectedDER("Trailing junk after DER public key")

    alg_id, rem = der.remove_sequence(s1)

    alg_id, rest = der.remove_object(alg_id)
    if alg_id not in OIDS:
        raise der.UnexpectedDER(f"Not recognised algoritm OID: {alg_id}")
    if rest:
        raise der.UnexpectedDER("Parameters specified for an ML-DSA OID")

    ml_dsa = OIDS[alg_id]

    key, empty = der.remove_bitstring(rem, 0)
    if empty:
        raise der.UnexpectedDER("Trailing junk after public key bitstring")
    if len(key) != ml_dsa._pk_size():
        raise der.UnexpectedDER("Wrong key size for the OID in structure")

    return ml_dsa, key


def pk_from_pem(enc_key):
    """
    Extract a verification key from PEM encoding.

    :param str enc_key: SPKI PEM encoding of a key
    :rtype: tuple(ML_DSA, bytes)
    """
    der_key = der.unpem(enc_key)
    return pk_from_der(der_key)


def sk_to_der(ml_dsa, sk=None, seed=None, form=None):
    """
    Convert a signing key to a PKCS #8 DER structure.

    ``sk``, ``seed``, or both need to be specified.
    if ``form`` is not specified (is set to ``None``). The format that
    preserves maximum amount of information will be used.

    Proposed in draft-ietf-lamps-dilithium-certificates-12

    :param ml_dsa: an ``ML_DSA`` object instance
    :param bytes sk: signing key
    :param bytes seed: seed to generate the ML_DSA keys
    :param str form: What format to write the key in, options are:
        None - for automatic selection based on ``sk`` and ``seed``,
        ``seed`` for writing seed only, ``expanded`` for writing the expanded
        key only, and ``both`` for writing both seed and expanded key.
    :rtype: bytes
    """
    if not ml_dsa.oid:
        raise ValueError("Only ML_DSAs with specified OIDs can be encoded")

    if form not in ("seed", "expanded", "both", None):
        raise ValueError(
            f"Invalid form specified: {form}. "
            "Only 'seed', 'expanded', 'both', or None are allowed"
        )

    if not sk and not seed:
        raise ValueError("sk or seed must be specified")

    if sk and len(sk) != ml_dsa._sk_size():
        raise ValueError("Invalid signing key size for the provided ML_DSA")

    if seed and len(seed) != 32:
        raise ValueError("Invalid seed size")

    if form in ("both", "seed") and not seed:
        raise ValueError(f'Format "{form}" requires specifying seed')

    if form is None:
        if sk and seed:
            form = "both"
        elif sk:
            form = "expanded"
        else:
            assert seed
            form = "seed"

    if form in ("both", "expanded") and not sk:
        _, sk = ml_dsa.key_derive(seed)

    if form == "seed":
        enc_key = der.encode_implicit(0, seed)
    elif form == "expanded":
        enc_key = der.encode_octet_string(sk)
    else:
        assert form == "both"
        enc_key = der.encode_sequence(
            der.encode_octet_string(seed), der.encode_octet_string(sk)
        )

    encoded_pkcs8 = der.encode_sequence(
        der.encode_integer(0),
        der.encode_sequence(der.encode_oid(*ml_dsa.oid)),
        der.encode_octet_string(enc_key),
    )

    return encoded_pkcs8


def sk_to_pem(ml_dsa, sk=None, seed=None, form=None):
    """
    Convert a signing key to a PKCS #8 PEM structure.

    ``sk``, ``seed``, or both need to be specified.
    if ``form`` is not specified (is set to ``None``). The format that
    preserves maximum amount of information will be used.

    Proposed in draft-ietf-lamps-dilithium-certificates-12

    :param ml_dsa: an ``ML_DSA`` object instance
    :param bytes sk: signing key
    :param bytes seed: seed to generate the ML_DSA keys
    :param str form: What format to write the key in, options are:
        None - for automatic selection based on ``sk`` and ``seed``,
        ``seed`` for writing seed only, ``expanded`` for writing the expanded
        key only, and ``both`` for writing both seed and expanded key.
    :rtype: bytes
    """
    der_enc = sk_to_der(ml_dsa, sk, seed, form)

    pem_enc = der.topem(der_enc, "PRIVATE KEY")

    return pem_enc


def sk_from_der(enc_key):
    """
    Extract signing and verification key from from PKCS #8 DER encoding.

    :param bytes enc_key: PKCS #8 DER encoding of the key.
    :return: the first element is the ``ML_DSA`` object instance, the second
        is the signing key, third is the seed (if present), fourth is the
        verification key (if seed is present).
    :rtype: tuple(ML_DSA, bytes, bytes, bytes)
    """
    s1, empty = der.remove_sequence(enc_key)
    if empty:
        raise der.UnexpectedDER("Trailing junk after private key structure")

    ver, rest = der.remove_integer(s1)
    if ver != 0:
        raise der.UnexpectedDER(f"Unsupported version: {ver}")

    alg_id, rest = der.remove_sequence(rest)

    alg_id, empty = der.remove_object(alg_id)
    if empty:
        raise der.UnexpectedDER("Junk after algorithm OID")
    if alg_id not in OIDS:
        raise der.UnexpectedDER(f"Not recognised algorithm OID: {alg_id}")

    ml_dsa = OIDS[alg_id]

    priv_key, _ = der.remove_octet_string(rest)
    # rest can be either parameters or public key: we ignore those

    seed = None
    expanded = None
    pk = None

    if der.str_idx_as_int(priv_key, 0) == 0x04:
        # we have OCTET STRING, expanded only format
        expanded, empty = der.remove_octet_string(priv_key)
        if empty:
            raise der.UnexpectedDER("Trailing junk after expandedKey")
    elif der.is_sequence(priv_key):
        both, empty = der.remove_sequence(priv_key)
        if empty:
            raise der.UnexpectedDER("Trailing junk after both encoding")
        seed, key_val = der.remove_octet_string(both)
        expanded, empty = der.remove_octet_string(key_val)
        if empty:
            raise der.UnexpectedDER(
                "Trailing junk after 'expandedKey' in 'both' encoding"
            )
    else:
        tag, seed, empty = der.remove_implicit(priv_key)
        if tag != 0:
            raise der.UnexpectedDER(f"Unexpected tag in private key encoding: {tag}")
        if empty:
            raise der.UnexpectedDER("Junk after seed encoding")

    if expanded and len(expanded) != ml_dsa._sk_size():
        raise der.UnexpectedDER("Invalid expanded key size in encoding")

    if seed and len(seed) != 32:
        raise der.UnexpectedDER("Invalid length of seed in encoding")

    if not expanded:
        pk, expanded = ml_dsa.key_derive(seed)

    if not pk:
        # If we reach here, we need to compute the public key
        # directly from the secret key bytes
        pk = ml_dsa.pk_from_sk(expanded)

    return ml_dsa, expanded, seed, pk


def sk_from_pem(enc_key):
    """
    Extract signing and verification key from from PKCS #8 PEM encoding.

    :param bytes enc_key: PKCS #8 PEM encoding of the key.
    :return: the first element is the ``ML_DSA`` object instance, the second
        is the signing key, third is the seed (if present), fourth is the
        verification key (if seed is present).
    :rtype: tuple(ML_DSA, bytes, bytes, bytes)
    """
    der_key = der.unpem(enc_key)
    return sk_from_der(der_key)



from .ml_dsa import ML_DSA
from .hash_ml_dsa import HashML_DSA


DEFAULT_PARAMETERS = {
    "ML_DSA_44": {
        "d": 13,  # number of bits dropped from t
        "tau": 39,  # number of ±1 in c
        "gamma_1": 131072,  # coefficient range of y: 2^17
        "gamma_2": 95232,  # low order rounding range: (q-1)/88
        "k": 4,  # Dimensions of A = (k, l)
        "l": 4,  # Dimensions of A = (k, l)
        "eta": 2,  # Private key range
        "omega": 80,  # Max number of ones in hint
        "c_tilde_bytes": 32,
        "oid": (2, 16, 840, 1, 101, 3, 4, 3, 17),
    },
    "ML_DSA_65": {
        "d": 13,  # number of bits dropped from t
        "tau": 49,  # number of ±1 in c
        "gamma_1": 524288,  # coefficient range of y: 2^19
        "gamma_2": 261888,  # low order rounding range: (q-1)/32
        "k": 6,  # Dimensions of A = (k, l)
        "l": 5,  # Dimensions of A = (k, l)
        "eta": 4,  # Private key range
        "omega": 55,  # Max number of ones in hint
        "c_tilde_bytes": 48,
        "oid": (2, 16, 840, 1, 101, 3, 4, 3, 18),
    },
    "ML_DSA_87": {
        "d": 13,  # number of bits dropped from t
        "tau": 60,  # number of ±1 in c
        "gamma_1": 524288,  # coefficient range of y: 2^19
        "gamma_2": 261888,  # low order rounding range: (q-1)/32
        "k": 8,  # Dimensions of A = (k, l)
        "l": 7,  # Dimensions of A = (k, l)
        "eta": 2,  # Private key range
        "omega": 75,  # Max number of ones in hint
        "c_tilde_bytes": 64,
        "oid": (2, 16, 840, 1, 101, 3, 4, 3, 19),
    },
    "HASH_ML_DSA_44_WITH_SHA512": {
        "d": 13,  # number of bits dropped from t
        "tau": 39,  # number of ±1 in c
        "gamma_1": 131072,  # coefficient range of y: 2^17
        "gamma_2": 95232,  # low order rounding range: (q-1)/88
        "k": 4,  # Dimensions of A = (k, l)
        "l": 4,  # Dimensions of A = (k, l)
        "eta": 2,  # Private key range
        "omega": 80,  # Max number of ones in hint
        "c_tilde_bytes": 32,
        "oid": (2, 16, 840, 1, 101, 3, 4, 3, 32),
    },
    "HASH_ML_DSA_65_WITH_SHA512": {
        "d": 13,  # number of bits dropped from t
        "tau": 49,  # number of ±1 in c
        "gamma_1": 524288,  # coefficient range of y: 2^19
        "gamma_2": 261888,  # low order rounding range: (q-1)/32
        "k": 6,  # Dimensions of A = (k, l)
        "l": 5,  # Dimensions of A = (k, l)
        "eta": 4,  # Private key range
        "omega": 55,  # Max number of ones in hint
        "c_tilde_bytes": 48,
        "oid": (2, 16, 840, 1, 101, 3, 4, 3, 33),
    },
    "HASH_ML_DSA_87_WITH_SHA512": {
        "d": 13,  # number of bits dropped from t
        "tau": 60,  # number of ±1 in c
        "gamma_1": 524288,  # coefficient range of y: 2^19
        "gamma_2": 261888,  # low order rounding range: (q-1)/32
        "k": 8,  # Dimensions of A = (k, l)
        "l": 7,  # Dimensions of A = (k, l)
        "eta": 2,  # Private key range
        "omega": 75,  # Max number of ones in hint
        "c_tilde_bytes": 64,
        "oid": (2, 16, 840, 1, 101, 3, 4, 3, 34),
    },
}

ML_DSA_44 = ML_DSA(DEFAULT_PARAMETERS["ML_DSA_44"])
ML_DSA_65 = ML_DSA(DEFAULT_PARAMETERS["ML_DSA_65"])
ML_DSA_87 = ML_DSA(DEFAULT_PARAMETERS["ML_DSA_87"])

HASH_ML_DSA_44_WITH_SHA512 = HashML_DSA(DEFAULT_PARAMETERS["ML_DSA_44"])
HASH_ML_DSA_65_WITH_SHA512 = HashML_DSA(DEFAULT_PARAMETERS["ML_DSA_65"])
HASH_ML_DSA_87_WITH_SHA512 = HashML_DSA(DEFAULT_PARAMETERS["ML_DSA_87"])



import unittest
from random import randint
from dilithium_py.polynomials.polynomials_generic import PolynomialRingGeneric


class TestPolynomialRing(unittest.TestCase):
    R = PolynomialRingGeneric(11, 5)

    def test_gen(self):
        self.assertTrue(self.R.gen() == self.R([0, 1]))

    def test_random_element(self):
        for _ in range(100):
            f = self.R.random_element()
            self.assertEqual(type(f), self.R.element)
            self.assertEqual(len(f.coeffs), self.R.n)
            self.assertTrue(all([c < self.R.q for c in f.coeffs]))

    def test_non_list_error(self):
        self.assertRaises(TypeError, lambda: self.R("1"))

    def test_long_list_error(self):
        self.assertRaises(ValueError, lambda: self.R([0] * (self.R.n + 1)))

    def test_string_format(self):
        self.assertEqual(
            str(self.R),
            "Univariate Polynomial Ring in x over Finite Field of size 11 with modulus x^5 + 1",
        )


class TestPolynomial(unittest.TestCase):
    R = PolynomialRingGeneric(11, 5)

    def test_getitem(self):
        x = self.R.gen()
        self.assertEqual(x[0], 0)
        self.assertEqual(x[1], 1)

    def test_is_zero(self):
        self.assertTrue(self.R(0).is_zero())
        self.assertFalse(self.R(1).is_zero())

    def test_is_constant(self):
        self.assertTrue(self.R(0).is_constant())
        self.assertTrue(self.R(1).is_constant())
        self.assertFalse(self.R.gen().is_constant())

    def test_reduce_coefficients(self):
        for _ in range(100):
            # Create non-canonical coefficients
            coeffs = [randint(-2 * self.R.q, 3 * self.R.q) for _ in range(self.R.n)]
            f = self.R(coeffs).reduce_coefficients()
            self.assertTrue(all([0 <= c < self.R.q for c in f.coeffs]))

    def test_equality(self):
        for _ in range(100):
            f1 = self.R.random_element()
            f2 = -f1
            self.assertEqual(f1, f1)
            # We don't cover the case of f1 being zero, as it's incredibly unlikely to happen
            if f1.is_zero():
                self.assertTrue(f1 == f2)  # pragma: no cover
            else:
                self.assertFalse(f1 == f2)

        self.assertTrue(self.R(0) == 0)
        self.assertTrue(self.R(1) == self.R.q + 1)
        self.assertTrue(self.R(self.R.q - 1) == -1)
        self.assertTrue(self.R(0) != 1)
        self.assertFalse(self.R(self.R.q - 1) == "a")

    def test_add_failure(self):
        f1 = self.R.random_element()
        self.assertRaises(NotImplementedError, lambda: f1 + "a")

    def test_sub_failure(self):
        f1 = self.R.random_element()
        self.assertRaises(NotImplementedError, lambda: f1 - "a")

    def test_mul_failure(self):
        f1 = self.R.random_element()
        self.assertRaises(NotImplementedError, lambda: f1 * "a")

    def test_pow_failure(self):
        f1 = self.R.random_element()
        self.assertRaises(TypeError, lambda: f1 ** "a")

    def test_add_polynomials(self):
        zero = self.R(0)
        for _ in range(100):
            f1 = self.R.random_element()
            f2 = self.R.random_element()
            f3 = self.R.random_element()

            self.assertEqual(f1 + zero, f1)
            self.assertEqual(f1 + f2, f2 + f1)
            self.assertEqual(f1 + (f2 + f3), (f1 + f2) + f3)

            f2 = f1
            f2 += f1
            self.assertEqual(f1 + f1, f2)

    def test_sub_polynomials(self):
        zero = self.R(0)
        for _ in range(100):
            f1 = self.R.random_element()
            f2 = self.R.random_element()
            f3 = self.R.random_element()

            self.assertEqual(f1 - zero, f1)
            self.assertEqual(f3 - f3, zero)
            self.assertEqual(f3 - 0, f3)
            self.assertEqual(0 - f3, -f3)
            self.assertEqual(f1 - f2, -(f2 - f1))
            self.assertEqual(f1 - (f2 - f3), (f1 - f2) + f3)

            f2 = f1
            f2 -= f1
            self.assertEqual(f2, zero)

    def test_mul_polynomials(self):
        zero = self.R(0)
        one = self.R(1)
        for _ in range(100):
            f1 = self.R.random_element()
            f2 = self.R.random_element()
            f3 = self.R.random_element()

            self.assertEqual(f1 * zero, zero)
            self.assertEqual(f1 * one, f1)
            self.assertEqual(f1 * f2, f2 * f1)
            self.assertEqual(f1 * (f2 * f3), (f1 * f2) * f3)
            self.assertEqual(2 * f1, f1 + f1)
            self.assertEqual(2 * f1, f1 * 2)

            f2 = f1
            f2 *= f2
            self.assertEqual(f1 * f1, f2)

    def test_pow_polynomials(self):
        one = self.R(1)
        for _ in range(100):
            f1 = self.R.random_element()

            self.assertEqual(one, f1**0)
            self.assertEqual(f1, f1**1)
            self.assertEqual(f1 * f1, f1**2)
            self.assertEqual(f1 * f1 * f1, f1**3)
            self.assertRaises(ValueError, lambda: f1 ** (-1))

    def test_print(self):
        self.assertEqual(str(self.R(0)), "0")
        self.assertEqual(str(self.R(1)), "1")
        self.assertEqual(str(self.R.gen()), "x")
        self.assertEqual(str(self.R([1, 2, 3, 4, 1])), "1 + 2*x + 3*x^2 + 4*x^3 + x^4")



from .dilithium import Dilithium

DEFAULT_PARAMETERS = {
    "dilithium2": {
        "d": 13,
        "k": 4,
        "l": 4,
        "eta": 2,
        "tau": 39,
        "omega": 80,
        "gamma_1": 131072,  # 2^17
        "gamma_2": 95232,  # (q-1)/88
    },
    "dilithium3": {
        "d": 13,
        "k": 6,
        "l": 5,
        "eta": 4,
        "tau": 49,
        "omega": 55,
        "gamma_1": 524288,  # 2^19
        "gamma_2": 261888,  # (q-1)/32
    },
    "dilithium5": {
        "d": 13,
        "k": 8,
        "l": 7,
        "eta": 2,
        "tau": 60,
        "omega": 75,
        "gamma_1": 524288,  # 2^19
        "gamma_2": 261888,  # (q-1)/32
    },
}

Dilithium2 = Dilithium(DEFAULT_PARAMETERS["dilithium2"])
Dilithium3 = Dilithium(DEFAULT_PARAMETERS["dilithium3"])
Dilithium5 = Dilithium(DEFAULT_PARAMETERS["dilithium5"])



from .ml_dsa import ML_DSA
from hashlib import sha256, sha512, shake_128


class HashML_DSA(ML_DSA):
    def _hash_with_oid(self, m: bytes, hash_name: str) -> tuple[bytes, bytes]:
        hash_name = hash_name.upper()

        if hash_name == "SHA256":
            oid = bytes(
                [0x06, 0x09, 0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x01]
            )
            ph_m = sha256(m).digest()
        elif hash_name == "SHA512":
            oid = bytes(
                [0x06, 0x09, 0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x03]
            )
            ph_m = sha512(m).digest()
        elif hash_name == "SHAKE128":
            oid = bytes(
                [0x06, 0x09, 0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x0B]
            )
            ph_m = shake_128(m).digest(32)
        else:
            raise ValueError(f"unsupported hash algorithm: {hash_name}")

        return oid, ph_m

    def _sign_with_pre_hash(
        self,
        sk: bytes,
        m: bytes,
        hash_name: str,
        ctx: bytes = b"",
        deterministic: bool = False,
    ) -> bytes:
        """
        Generates a HashML-DSA signature following Algorithm 4 (FIPS 204)

        The hash name is a string which selects the pre-hash function and
        can currently be SHA256, SHA512 or SHAKE128.
        """
        if len(ctx) > 255:
            raise ValueError(
                f"ctx bytes must have length at most 255, ctx has length {len(ctx) = }"
            )

        if deterministic:
            rnd = bytes([0] * 32)
        else:
            rnd = self.random_bytes(32)

        # Prehash the message and return the OID of the used hash
        oid, ph_m = self._hash_with_oid(m, hash_name)

        # Format the message using the context
        m_prime = bytes([1]) + bytes([len(ctx)]) + ctx + oid + ph_m

        # Compute the signature of m_prime
        sig_bytes = self._sign_internal(sk, m_prime, rnd)
        return sig_bytes

    def _verify_with_pre_hash(
        self, pk: bytes, m: bytes, sig: bytes, hash_name: str, ctx: bytes = b""
    ) -> bool:
        """
        Verifies a signature sigma for a message M following algorithm 5 (FIPS 204)
        """
        if len(ctx) > 255:
            raise ValueError(
                f"ctx bytes must have length at most 255, ctx has length {len(ctx) = }"
            )

        # Prehash the message and return the OID of the used hash
        oid, ph_m = self._hash_with_oid(m, hash_name)

        # Format the message using the context
        m_prime = bytes([1]) + bytes([len(ctx)]) + ctx + oid + ph_m

        return self._verify_internal(pk, m_prime, sig)

    def sign(
        self,
        sk: bytes,
        m: bytes,
        ctx: bytes = b"",
        deterministic: bool = False,
    ) -> bytes:
        """
        Generates a HashML-DSA signature following Algorithm 4 (FIPS 204)
        with SHA512 as the chosen hash function.
        """
        return self._sign_with_pre_hash(sk, m, "SHA512", ctx, deterministic)

    def verify(self, pk: bytes, m: bytes, sig: bytes, ctx: bytes = b"") -> bool:
        """
        Verifies a signature sigma for a message M following algorithm 5 (FIPS 204)
        with SHA512 as the chosen hash function.
        """
        return self._verify_with_pre_hash(pk, m, sig, "SHA512", ctx)



import unittest
import os
from dilithium_py.ml_dsa import (
    HASH_ML_DSA_44_WITH_SHA512,
    HASH_ML_DSA_65_WITH_SHA512,
    HASH_ML_DSA_87_WITH_SHA512,
)


class TestHashMLDSA(unittest.TestCase):
    """
    Test ML DSA for internal consistency by generating signatures
    and verifying them!
    """

    def generic_test_hash_ml_dsa(self, HASH_ML_DSA, hash_name="SHA512", count=5):
        for _ in range(count):
            msg = b"Signed by HASH_ML_DSA" + os.urandom(16)
            ctx = os.urandom(128)

            # Perform signature process
            pk, sk = HASH_ML_DSA.keygen()
            sig = HASH_ML_DSA.sign(sk, msg, ctx=ctx)
            check_verify = HASH_ML_DSA.verify(pk, msg, sig, ctx=ctx)

            # Generate some fail cases
            pk_bad, _ = HASH_ML_DSA.keygen()
            check_wrong_pk = HASH_ML_DSA.verify(pk_bad, msg, sig, ctx=ctx)
            check_wrong_msg = HASH_ML_DSA.verify(pk, b"", sig, ctx=ctx)
            check_no_ctx = HASH_ML_DSA.verify(pk, msg, sig)

            # Check with user-supplied hashes
            hash_sig = HASH_ML_DSA._sign_with_pre_hash(sk, msg, hash_name, ctx=ctx)
            check_hash_verify = HASH_ML_DSA._verify_with_pre_hash(
                pk, msg, hash_sig, hash_name, ctx=ctx
            )

            # Check with the wrong hashes
            if hash_name == "SHA512":
                bad_hash = "SHA256"
            else:
                bad_hash = "SHA512"
            check_wrong_hash = HASH_ML_DSA._verify_with_pre_hash(
                pk, msg, hash_sig, bad_hash, ctx=ctx
            )

            # Check that signature works
            self.assertTrue(check_verify)

            # Check that signature works with custom hash
            self.assertTrue(check_hash_verify)

            # Ensure the hashes need to match
            self.assertFalse(check_wrong_hash)

            # Check changing the key breaks verify
            self.assertFalse(check_wrong_pk)

            # Check changing the message breaks verify
            self.assertFalse(check_wrong_msg)

            # Check removing the context breaks verify
            self.assertFalse(check_no_ctx)

    # Default hash is SHA512
    def test_hash_ml_dsa_44(self):
        self.generic_test_hash_ml_dsa(HASH_ML_DSA_44_WITH_SHA512)

    def test_hash_ml_dsa_65(self):
        self.generic_test_hash_ml_dsa(HASH_ML_DSA_65_WITH_SHA512)

    def test_hash_ml_dsa_87(self):
        self.generic_test_hash_ml_dsa(HASH_ML_DSA_87_WITH_SHA512)

    # Test with SHA256
    def test_hash_ml_dsa_44_sha256(self):
        self.generic_test_hash_ml_dsa(HASH_ML_DSA_44_WITH_SHA512, "SHA256")

    def test_hash_ml_dsa_65_sha256(self):
        self.generic_test_hash_ml_dsa(HASH_ML_DSA_65_WITH_SHA512, "SHA256")

    def test_hash_ml_dsa_87_sha256(self):
        self.generic_test_hash_ml_dsa(HASH_ML_DSA_87_WITH_SHA512, "SHA256")

    # Test with SHAKE128
    def test_hash_ml_dsa_44_shake128(self):
        self.generic_test_hash_ml_dsa(HASH_ML_DSA_44_WITH_SHA512, "SHAKE128")

    def test_hash_ml_dsa_65_shake128(self):
        self.generic_test_hash_ml_dsa(HASH_ML_DSA_65_WITH_SHA512, "SHAKE128")

    def test_hash_ml_dsa_87_shake128(self):
        self.generic_test_hash_ml_dsa(HASH_ML_DSA_87_WITH_SHA512, "SHAKE128")


class TestHashMLDSADeterministic(unittest.TestCase):
    """
    Test ML DSA for internal consistency by generating signatures
    and verifying them!
    """

    def generic_test_hash_ml_dsa(self, HASH_ML_DSA, count=5):
        for _ in range(count):
            msg = b"Signed by HASH_ML_DSA" + os.urandom(16)
            ctx = os.urandom(128)

            # Perform signature process
            pk, sk = HASH_ML_DSA.keygen()
            sig = HASH_ML_DSA.sign(sk, msg, ctx=ctx, deterministic=True)
            check_verify = HASH_ML_DSA.verify(pk, msg, sig, ctx=ctx)

            # Generate some fail cases
            pk_bad, _ = HASH_ML_DSA.keygen()
            check_wrong_pk = HASH_ML_DSA.verify(pk_bad, msg, sig, ctx=ctx)
            check_wrong_msg = HASH_ML_DSA.verify(pk, b"", sig, ctx=ctx)
            check_no_ctx = HASH_ML_DSA.verify(pk, msg, sig)

            # Check that signature works
            self.assertTrue(check_verify)

            # Check changing the key breaks verify
            self.assertFalse(check_wrong_pk)

            # Check changing the message breaks verify
            self.assertFalse(check_wrong_msg)

            # Check removing the context breaks verify
            self.assertFalse(check_no_ctx)

    def test_hash_ml_dsa_44(self):
        self.generic_test_hash_ml_dsa(HASH_ML_DSA_44_WITH_SHA512)

    def test_hash_ml_dsa_65(self):
        self.generic_test_hash_ml_dsa(HASH_ML_DSA_65_WITH_SHA512)

    def test_hash_ml_dsa_87(self):
        self.generic_test_hash_ml_dsa(HASH_ML_DSA_87_WITH_SHA512)



def reduce_mod_pm(r, n):
    """
    Takes an integer 0 < r < n and computes the value
    r' = r mod^± n, defined to be integer in the range

    for n even:
        -(n / 2) < r' <= (n / 2)
    for n odd:
        -(n - 1) / 2 <= r' <= (n - 1) / 2
    """
    r = r % n
    if r > (n >> 1):
        r -= n
    return r


def decompose(r, a, q):
    """
    Takes an element r and represents
    it as:

    r = r1*a + r0

    With r0 in the range

    -(a << 1) < r0 <= (a << 1)
    """
    rp = r % q
    r0 = reduce_mod_pm(rp, a)
    r1 = rp - r0
    if rp - r0 == q - 1:
        r1 = 0
        r0 = r0 - 1
    else:
        r1 = (rp - r0) // a

    # assert r0 > -(a >> 1)
    # assert r0 <= (a >> 1)
    # assert r % q == (r0 + r1 * a) % q
    return r1, r0


def high_bits(r, a, q):
    r1, _ = decompose(r, a, q)
    return r1


def low_bits(r, a, q):
    _, r0 = decompose(r, a, q)
    return r0


def make_hint(z, r, a, q):
    """
    Check whether the top bit of z will change when r is added
    """
    r1 = high_bits(r, a, q)
    v1 = high_bits(r + z, a, q)
    return int(r1 != v1)


def make_hint_optimised(z0, r1, a, q):
    """
    Optimised version of the above used when the low bits w0 are extracted from
    `w = (A_hat @ y_hat).from_ntt()` during signing
    """
    gamma2 = a >> 1
    if z0 <= gamma2 or z0 > (q - gamma2) or (z0 == (q - gamma2) and r1 == 0):
        return 0
    return 1


def use_hint(h, r, a, q):
    m = (q - 1) // a
    r1, r0 = decompose(r, a, q)
    if h == 1:
        if r0 > 0:
            return (r1 + 1) % m
        return (r1 - 1) % m
    return r1


def check_norm_bound(n, b, q):
    """
    Norm bound is checked in the following four steps:
    x ∈ {0,        ...,                    ...,     q-1}
    x ∈ {-(q-1)/2, ...,       -1,       0, ..., (q-1)/2}
    x ∈ { (q-3)/2, ...,        0,       0, ..., (q-1)/2}
    x ∈ {0, 1,     ...,  (q-1)/2, (q-1)/2, ...,       1}
    """
    x = n % q
    x = ((q - 1) >> 1) - x
    x = x ^ (x >> 31)
    x = ((q - 1) >> 1) - x
    return x >= b


def xor_bytes(a, b):
    """
    XOR two byte arrays, assume that they are
    of the same length
    """
    return bytes(a ^ b for a, b in zip(a, b))



from dilithium_py.ml_dsa import ML_DSA_44, ML_DSA_65, ML_DSA_87
import cProfile
from time import time
from statistics import mean, median


def profile_ml_dsa(ML_DSA):
    pk, sk = ML_DSA.keygen()
    m = b"Signed by ml_dsa"
    sig = ML_DSA.sign(sk, m)
    check = ML_DSA.verify(pk, m, sig)
    assert check

    gvars = {}
    lvars = {"ML_DSA": ML_DSA, "m": m, "pk": pk, "sk": sk, "sig": sig}

    cProfile.runctx(
        "[ML_DSA.keygen() for _ in range(500)]", globals=gvars, locals=lvars, sort=1
    )
    # cProfile.runctx(
    #     "[ML_DSA.sign(sk, m) for _ in range(500)]",
    #     globals=gvars,
    #     locals=lvars,
    #     sort=1,
    # )
    # cProfile.runctx(
    #     "[ML_DSA.verify(pk, m, sig) for _ in range(500)]",
    #     globals=gvars,
    #     locals=lvars,
    #     sort=1,
    # )


def benchmark_ml_dsa(ML_DSA, name, count):
    # Banner
    print("-" * 27)
    print(f"  {name} | ({count} calls)")
    print("-" * 27)

    fails = 0
    keygen_times = []
    sign_times = []
    verify_times = []
    # 32 byte message
    m = b"Your message signed by ML_DSA"

    for _ in range(count):
        t0 = time()
        pk, sk = ML_DSA.keygen()
        keygen_times.append(time() - t0)

        t1 = time()
        sig = ML_DSA.sign(sk, m)
        sign_times.append(time() - t1)

        t2 = time()
        verify = ML_DSA.verify(pk, m, sig)
        verify_times.append(time() - t2)
        if not verify:
            fails += 1

    print(f"Keygen median: {round(median(keygen_times), 3)}")
    print(f"Sign median: {round(median(sign_times), 3)}")
    print(f"Sign average: {round(mean(sign_times), 3)}")
    print(f"Verify median: {round(median(verify_times), 3)}")
    print(f"Fails: {fails}")


if __name__ == "__main__":
    # I used 1000 calls for the README, but you might want to
    # shrink this down if you're experimenting
    count = 1000
    benchmark_ml_dsa(ML_DSA_44, "ML_DSA_44", count)
    benchmark_ml_dsa(ML_DSA_65, "ML_DSA_65", count)
    benchmark_ml_dsa(ML_DSA_87, "ML_DSA_87", count)

    # profile_ml_dsa(ML_DSA_44)
